#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
Contains classes and functions to create and maintain file structures
for uploads, and some generic file utilities.

There are two different structures for uploads in two different states: *staging* and *public*.
Possible operations on uploads differ based on this state. Staging is used for
processing, heavily editing, creating hashes, etc. Public is supposed to be a
almost readonly (beside metadata) storage.

.. code-block:: sh

    fs/staging/<upload>/raw/**
                       /archive/<entry_id>.msg
    fs/public/<upload>/raw-{access}.plain.zip
                      /archive-{access}.msg.msg

Where `access` is either "public" (non-embargoed) or "restricted" (embargoed).

There is an implicit relationship between files, based on them being in the same
directory. Each directory with at least one *mainfile* is an *entry directory*
and all the files are *aux* files to that mainfile. This is independent of whether the
respective files actually contributes data or not. An entry directory might
contain multiple mainfiles. E.g., user simulated multiple states of the same system, have
one entry based on the other, etc. In this case the other mainfile is an *aux file* to the
original mainfile, and vice versa.
'''

from abc import ABCMeta
import sys
from typing import IO, Dict, Iterable, Iterator, List, Tuple, Any, NamedTuple
from functools import lru_cache
from pydantic import BaseModel
from datetime import datetime
import os.path
import os
import shutil
import tarfile
import zipstream
import hashlib
import io
import json
import yaml
import magic

from nomad import config, utils, datamodel
from nomad.archive import write_archive, read_archive, ArchiveReader

# TODO this should become obsolete, once we are going beyong python 3.6. For now
# python 3.6's zipfile does not allow to seek/tell within a file-like opened from a
# file in a zipfile.
if sys.version_info >= (3, 7):
    import zipfile
else:
    import zipfile37 as zipfile

zip_file_extensions = ('.zip',)
tar_file_extensions = ('.tgz', '.gz', '.tar.gz', '.tar.bz2', '.tar')
bundle_info_filename = 'bundle_info.json'

# Used to check if zip-files/archive files are empty
# TODO: These should not be needed when we move on to only keep files with the right access
empty_zip_file_size = 22
empty_archive_file_size = 32


def copytree(src, dst):
    '''
    A close on ``shutils.copytree`` that does not try to copy the stats on all files.
    This is unecessary for our usecase and also causes permission denies for unknown
    reasons.
    '''
    os.makedirs(dst, exist_ok=False)

    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            copytree(s, d)
        else:
            shutil.copyfile(s, d)


def create_tmp_dir(prefix: str) -> str:
    '''
    Creates a temporary directory in the directory specified by `config.fs.tmp`. The name
    of the directory will first be set to `prefix`, but if that name is already taken, a
    suffix will be added to ensure a completely clean, new directory is created. If prefix
    contains a '/', it will be replaced with '_', to ensure the validity of the path.
    The full path to the created directory is returned.
    '''
    assert prefix
    prefix = prefix.replace(os.path.sep, '_')
    assert is_safe_basename(prefix)
    for index in range(1, 100):
        dir_name = prefix if index == 1 else f'{prefix}_{index}'
        path = os.path.join(config.fs.tmp, dir_name)
        try:
            os.makedirs(path)
            return path
        except FileExistsError:
            pass  # Try again with different suffix
    raise RuntimeError('Could not create temporary directory - too many directories with same prefix?')


def is_safe_basename(basename: str) -> bool:
    '''
    Checks if `basename` is a *safe* base name (file/folder name). We consider it safe if
    it is not empty, does not contain any '/', and is not equal to '.' or '..'
    '''
    if not basename or '/' in basename or basename == '.' or basename == '..':
        return False
    return True


def is_safe_relative_path(path: str) -> bool:
    '''
    Checks if path is a *safe* relative path. We consider it safe if it does not start with
    '/' or use '.' or '..' elements (which could be open for security leaks if allowed).
    It may end with a single '/', indicating that a folder is referred. For referring to
    the base folder, the empty string should be used (not '.' etc).
    '''
    if type(path) != str:
        return False
    if path == '':
        return True
    if path.startswith('/') or '//' in path or '\n' in path:
        return False
    for element in path.split('/'):
        if element == '.' or element == '..':
            return False
    return True


class PathObject:
    '''
    Object storage-like abstraction for paths in general.
    Attributes:
        os_path: The full os path of the object.
    '''
    def __init__(self, os_path: str):
        self.os_path = os_path

    def delete(self) -> None:
        if os.path.isfile(self.os_path):
            os.remove(self.os_path)
        else:
            shutil.rmtree(self.os_path)

    def exists(self) -> bool:
        return os.path.exists(self.os_path)

    @property
    def size(self) -> int:
        ''' The os determined file size. '''
        return os.stat(self.os_path).st_size

    def __repr__(self) -> str:
        return self.os_path


class DirectoryObject(PathObject):
    '''
    Object storage-like abstraction for directories.
    '''
    def __init__(self, os_path: str, create: bool = False):
        self.os_path = os_path
        if create and not os.path.isdir(self.os_path):
            os.makedirs(self.os_path)

    def join_dir(self, path, create: bool = False) -> 'DirectoryObject':
        return DirectoryObject(os.path.join(self.os_path, path), create)

    def join_file(self, path, create_dir: bool = False) -> PathObject:
        if create_dir:
            dirname = os.path.dirname(path)
            if dirname:
                dir_os_path = os.path.join(self.os_path, dirname)
                if not os.path.exists(dir_os_path):
                    os.makedirs(dir_os_path)
        return PathObject(os.path.join(self.os_path, path))

    def exists(self) -> bool:
        return os.path.isdir(self.os_path)


class RawPathInfo(NamedTuple):
    '''
    Stores basic info about a file or folder located at a specific raw path.
    '''
    path: str
    is_file: bool
    size: int
    access: str


class StreamedFile(BaseModel):
    '''
    Convenience class for representing a streamed file, together with information about
    file size and an associated path.
    '''
    f: Any
    path: str
    size: int


class FileSource(metaclass=ABCMeta):
    '''
    An abstract class which represents a generic "file source", from which some number of files
    can be retrieved. There are several different ways to create a file source, see subclasses.
    The files in the source are associated with paths and have known sizes.
    '''
    def to_streamed_files(self) -> Iterable[StreamedFile]:
        ''' Retrieves the files in the source as :class:`StreamedFile` objects. '''
        raise NotImplementedError()

    def to_zipstream(self) -> Iterator[bytes]:
        ''' Returns a zip stream with the files from this FileSource. '''
        return create_zipstream(self.to_streamed_files())

    def to_zipfile(self, path, overwrite: bool = False):
        '''
        Generates a zip file from the files in this FileSource and stores it to disk. The
        zipfile content is created by calling :func:`to_zipstream`.
        '''
        assert not os.path.isdir(path), 'Exporting to zip file requires a file path, not directory.'
        assert overwrite or not os.path.exists(path), '`path` already exists. Use `overwrite` to overwrite.'
        with open(path, 'wb') as f:
            for chunk in self.to_zipstream():
                f.write(chunk)

    def to_disk(self, destination_dir: str, move_files: bool = False, overwrite: bool = False):
        '''
        Writes the files from this FileSource to disk, uncompressed. The default implementation
        makes use of :func:`to_streamed_files`. The `destination_dir` should be a directory
        (it will be created if it does not exist). The `move_files` argument instructs
        the method to move the source files if possible.
        '''
        if not os.path.exists(destination_dir):
            os.makedirs(destination_dir)
        assert os.path.isdir(destination_dir), '`destination_dir` is not a directory'
        for streamed_file in self.to_streamed_files():
            assert is_safe_relative_path(streamed_file.path), 'Unsafe relative path encountered'
            os_path = os.path.join(destination_dir, streamed_file.path)
            dir_path = os.path.dirname(os_path)
            if os.path.exists(os_path):
                assert overwrite, 'Target already exists and `overwrite` is False'
                PathObject(os_path).delete()
            os.makedirs(dir_path, exist_ok=True)
            with open(os_path, 'wb') as output_file:
                for chunk in streamed_file.f:
                    output_file.write(chunk)

    def close(self):
        ''' Perform "closing" of the source, if applicable. '''
        pass


class BrowsableFileSource(FileSource, metaclass=ABCMeta):
    '''
    A :class:`FileSource` which can be "browsed", like a folder on disk or a zip archive.
    '''
    def open(self, path, mode='rb') -> IO:
        ''' Opens a file by the specified path. '''
        raise NotImplementedError()

    def directory_list(self, path: str) -> List[str]:
        '''
        Returns a list of directory contents, located in the directory denoted by `path`
        in this file source.
        '''
        raise NotImplementedError()

    def sub_source(self, path: str) -> 'BrowsableFileSource':
        '''
        Creates a new instance of :class:`BrowsableFileSource` which just contains the
        files located under the specified path.
        '''
        raise NotImplementedError()


class StreamedFileSource(FileSource):
    '''
    A :class:`FileSource` created from a single :class:`StreamedFile`.
    '''
    def __init__(self, streamed_file: StreamedFile):
        self.streamed_file = streamed_file

    def to_streamed_files(self) -> Iterable[StreamedFile]:
        yield self.streamed_file


class DiskFileSource(BrowsableFileSource):
    '''
    A :class:`FileSource` corresponding to a single file or a folder on disk. The object
    is identified by a `base_path` and a `relative path`. The `base_path` should be a folder,
    the `relative_path` is optional, and used for selecting only a specific file or folder
    located under `base_folder`. The paths of the files retrieved from this source are given
    relative to the `base_path`.
    '''
    def __init__(self, base_path: str, relative_path: str = None):
        assert os.path.isdir(base_path)
        if relative_path:
            assert is_safe_relative_path(relative_path), 'Unsafe relative_path received'
            self.full_path = os.path.join(base_path, relative_path)
            assert os.path.exists(self.full_path)
        else:
            self.full_path = base_path
        self.base_path = base_path
        self.relative_path = relative_path

    def to_streamed_files(self) -> Iterable[StreamedFile]:
        if os.path.isfile(self.full_path):
            # Single file
            yield StreamedFile(
                path=self.relative_path,
                f=open(self.full_path, 'rb'),
                size=os.stat(self.full_path).st_size)
        else:
            # Directory - crawl it and its subfolders for files
            for dirpath, __, filenames in os.walk(self.full_path):
                for filename in filenames:
                    sub_full_path = os.path.join(dirpath, filename)
                    sub_relative_path = os.path.relpath(sub_full_path, self.base_path)
                    yield StreamedFile(
                        path=sub_relative_path,
                        f=open(sub_full_path, 'rb'),
                        size=os.stat(sub_full_path).st_size)

    def to_disk(self, destination_dir: str, move_files: bool = False, overwrite: bool = False):
        if self.relative_path:
            destination_path = os.path.join(destination_dir, self.relative_path)
        else:
            destination_path = destination_dir
        destination_parent = os.path.dirname(destination_path)
        os.makedirs(destination_parent, exist_ok=True)
        if os.path.exists(destination_path):
            assert overwrite, f'Target {destination_path} already exists and `overwrite` is False'
            PathObject(destination_path).delete()
        # All looks good. Copy or move the source to the destination
        if move_files:
            shutil.move(self.full_path, destination_path)
        else:
            if os.path.isfile(self.full_path):
                shutil.copyfile(self.full_path, destination_path)
            else:
                copytree(self.full_path, destination_path)

    def open(self, path, mode='rb') -> IO:
        assert is_safe_relative_path(path)
        return open(os.path.join(self.base_path, path), mode)

    def directory_list(self, path: str) -> List[str]:
        assert is_safe_relative_path(path)
        sub_path = os.path.join(self.base_path, path)
        return os.listdir(sub_path)

    def sub_source(self, path: str) -> 'DiskFileSource':
        assert is_safe_relative_path(path)
        return DiskFileSource(self.base_path, path)


class ZipFileSource(BrowsableFileSource):
    '''
    Allows us to "wrap" a :class:`zipfile.ZipFile` object and use it as a :class:`BrowsableFileSource`,
    i.e. it denotes a resource (single file or folder) stored in a ZipFile.
    '''
    def __init__(self, zip_file: zipfile.ZipFile, sub_path: str = ''):
        assert is_safe_relative_path(sub_path)
        self.zip_file = zip_file
        self.sub_path = sub_path
        self._namelist: List[str] = zip_file.namelist()

    def to_streamed_files(self) -> Iterable[StreamedFile]:
        path_prefix = '' if not self.sub_path else self.sub_path + os.path.sep
        for path in self._namelist:
            if path == self.sub_path or (path.startswith(path_prefix) and not path.endswith(os.path.sep)):
                yield StreamedFile(
                    path=path,
                    f=self.zip_file.open(path, 'r'),
                    size=self.zip_file.getinfo(path).file_size)

    def open(self, path, mode='rb') -> IO:
        assert 'r' in mode, 'Mode must be a read mode'
        for c in mode:
            assert c in ('r', 'b', 't'), f'Invalid mode for open command: {mode}'
        f = self.zip_file.open(path, 'r')
        if 't' in mode:
            return io.TextIOWrapper(f)
        return f

    def directory_list(self, path: str) -> List[str]:
        path_prefix = '' if not path else path + os.path.sep
        found = set()
        for path2 in self._namelist:
            if path2.startswith(path_prefix):
                found.add(path2.split(os.path.sep)[0])
        return sorted(found)

    def sub_source(self, path: str) -> 'ZipFileSource':
        assert is_safe_relative_path(path), 'Unsafe path provided'
        if self.sub_path:
            assert path.startswith(self.sub_path + os.path.sep), 'Provided `path` is not a sub path.'
        return ZipFileSource(self.zip_file, path)

    def close(self):
        self.zip_file.close()


class CombinedFileSource(FileSource):
    '''
    Class for defining a :class:`FileSource` by combining multiple "subsources" into one.
    New sources are added using :func:`add_file_source`.
    '''
    def __init__(self):
        self.sources = []

    def add_file_source(self, file_source: FileSource):
        assert isinstance(file_source, FileSource)
        self.sources.append(file_source)

    def to_streamed_files(self) -> Iterable[StreamedFile]:
        for file_source in self.sources:
            for streamed_file in file_source.to_streamed_files():
                yield streamed_file

    def to_disk(self, destination_dir: str, move_files: bool = False, overwrite: bool = False):
        for file_source in self.sources:
            file_source.to_disk(destination_dir, move_files, overwrite)


class StandardJSONEncoder(json.JSONEncoder):
    """ Our standard JSONEncoder with support for marshalling of datetime objects """
    def default(self, obj):  # pylint: disable=E0202
        if isinstance(obj, datetime):
            return {'$datetime': obj.isoformat()}
        return json.JSONEncoder.default(self, obj)


class StandardJSONDecoder(json.JSONDecoder):
    """ Our standard JSONDecoder, with support for marshalling of datetime objects """
    def __init__(self, *args, **kargs):
        json.JSONDecoder.__init__(self, object_hook=self.dict_to_object, *args, **kargs)

    def dict_to_object(self, d):
        v = d.get('$datetime')
        if v is not None and len(d) == 1:
            return datetime.fromisoformat(v)
        return d


def json_to_streamed_file(json_dict: Dict[str, Any], path: str) -> StreamedFile:
    ''' Converts a json dictionary structure to a :class:`StreamedFile`. '''
    json_bytes = json.dumps(json_dict, indent=2, cls=StandardJSONEncoder).encode()
    return StreamedFile(
        path=path,
        f=io.BytesIO(json_bytes),
        size=len(json_bytes))


def create_zipstream_content(streamed_files: Iterable[StreamedFile]) -> Iterable[Dict]:
    '''
    Generator which "casts" a sequence of StreamedFiles to a sequence of dictionaries, of
    the form which is required by the `zipstream` library, i.e. dictionaries with keys
    `arcname`, `iterable` and `buffer_size`. Useful for generating zipstreams.
    '''
    for streamed_file in streamed_files:

        def content_generator():
            while True:
                data = streamed_file.f.read(1024 * 64)
                if not data:
                    break
                yield data

        yield dict(
            arcname=streamed_file.path,
            iterable=content_generator(),
            buffer_size=streamed_file.size)


def create_zipstream(
        streamed_files: Iterable[StreamedFile],
        compress: bool = False) -> Iterator[bytes]:
    '''
    Creates a zip stream, i.e. a streamed zip file.
    '''
    compression = zipfile.ZIP_DEFLATED if compress else zipfile.ZIP_STORED
    zip_stream = zipstream.ZipFile(mode='w', compression=compression, allowZip64=True)
    zip_stream.paths_to_write = create_zipstream_content(streamed_files)

    return iter(zip_stream)


class UploadFiles(DirectoryObject, metaclass=ABCMeta):
    ''' Abstract base class for upload files. '''
    def __init__(self, upload_id: str, create: bool = False):
        self.logger = utils.get_logger(__name__, upload_id=upload_id)

        super().__init__(os_path=self.base_folder_for(upload_id), create=create)

        if not create and not self.exists():
            raise KeyError(upload_id)

        self.upload_id = upload_id

    @classmethod
    def file_area(cls):
        '''
        Full path to where the upload files of this class are stored (i.e. either
        staging or public file area).
        '''
        raise NotImplementedError()

    @property
    def external_os_path(self):
        '''
        Full path to where the upload files of this class are stored on the server.
        This is equal to `self.os_path` if no external path substitutes for staging
        and public area are configured. This is helpful, when nomad is run in a container
        and the mounted path used by nomad are different from the actual paths on the
        host server.
        '''
        raise NotImplementedError()

    @classmethod
    def base_folder_for(cls, upload_id: str) -> str:
        '''
        Full path to the base folder for the upload files (of this class) for the
        specified upload_id.
        '''
        os_path = cls.file_area()
        if config.fs.prefix_size:
            os_path = os.path.join(os_path, upload_id[:config.fs.prefix_size])
        os_path = os.path.join(os_path, upload_id)
        return os_path

    @classmethod
    def exists_for(cls, upload_id: str) -> bool:
        '''
        If an UploadFiles object (of this class) has been created for this upload_id.
        '''
        return os.path.exists(cls.base_folder_for(upload_id))

    def to_staging_upload_files(self, create: bool = False, include_archive: bool = False) -> 'StagingUploadFiles':
        ''' Casts to or creates corresponding staging upload files or returns None. '''
        raise NotImplementedError()

    @staticmethod
    def get(upload_id: str, create: bool = False) -> 'UploadFiles':
        if StagingUploadFiles.exists_for(upload_id):
            return StagingUploadFiles(upload_id, create)
        elif PublicUploadFiles.exists_for(upload_id):
            return PublicUploadFiles(upload_id, create)
        else:
            return None

    def is_empty(self) -> bool:
        ''' If this upload has no content yet. '''
        raise NotImplementedError()

    def raw_path_exists(self, path: str) -> bool:
        '''
        Returns True if the specified path is a valid raw path (either file or directory)
        '''
        raise NotImplementedError()

    def raw_path_is_file(self, path: str) -> bool:
        '''
        Returns True if the specified path points to a file (rather than a directory).
        '''
        raise NotImplementedError()

    def raw_directory_list(
            self, path: str = '', recursive=False, files_only=False, path_prefix=None) -> Iterable[RawPathInfo]:
        '''
        Returns an iterable of RawPathInfo, one for each element (file or folder) in
        the directory specified by `path`. If `recursive` is set to True, subdirectories are
        also crawled. If `files_only` is set, only the file objects found are returned.
        If path is not a valid directory, the result will be empty. Selecting empty string
        as path (which is the default value) gives the content of the whole raw directory.
        The `path_prefix` argument can be used to filter out elements where the path starts
        with a specific prefix.
        '''
        raise NotImplementedError()

    def raw_file(self, file_path: str, *args, **kwargs) -> IO:
        '''
        Opens a raw file and returns a file-like object. Additional args, kwargs are
        delegated to the respective `open` call.
        Arguments:
            file_path: The path to the file relative to the upload.
        Raises:
            KeyError: If the file does not exist.
        '''
        raise NotImplementedError()

    def raw_file_size(self, file_path: str) -> int:
        '''
        Returns:
            The size of the given raw file.
        '''
        raise NotImplementedError()

    def raw_file_mime_type(self, file_path: str) -> str:
        assert self.raw_path_is_file(file_path), 'Provided path does not specify a file, or is invalid.'
        raw_file = self.raw_file(file_path, 'br')
        buffer = raw_file.read(2048)
        mime_type = magic.from_buffer(buffer, mime=True)
        raw_file.close()
        if not mime_type:
            mime_type = 'application/octet-stream'
        return mime_type

    def read_archive(self, entry_id: str) -> ArchiveReader:
        '''
        Returns an :class:`nomad.archive.ArchiveReader` that contains the
        given entry_id.
        '''
        raise NotImplementedError()

    def close(self):
        ''' Release possibly held system resources (e.g. file handles). '''
        pass

    def delete(self) -> None:
        shutil.rmtree(self.os_path)
        if config.fs.prefix_size > 0:
            # If using prefix, also remove the parent directory if empty
            parent_directory = os.path.dirname(self.os_path)
            if not os.listdir(parent_directory):
                try:
                    os.rmdir(parent_directory)
                except Exception as e:
                    utils.get_logger(__name__).error(
                        'could not remove empty prefix dir', directory=parent_directory, exc_info=e)

    def files_to_bundle(
            self, bundle_info: Dict[str, Any],
            include_raw_files: bool,
            include_archive_files: bool) -> FileSource:
        '''
        Returns a :class:`FileSource`, defining the files/folders to be included in an
        upload bundle when *exporting*. Note, the bundle_info.json file is not included,
        only the "regular" files, and only those specified by the arguments to the method.
        '''
        raise NotImplementedError()

    @classmethod
    def files_from_bundle(
            cls, bundle_file_source: BrowsableFileSource,
            include_raw_files: bool,
            include_archive_files: bool,
            include_bundle_info: bool) -> FileSource:
        '''
        Returns a :class:`FileSource`, defining the files/folders to be included in an
        upload bundle when *importing*. Only the files specified by the arguments are included.
        '''
        raise NotImplementedError()


class StagingUploadFiles(UploadFiles):
    def __init__(self, upload_id: str, create: bool = False):
        super().__init__(upload_id, create)

        self._raw_dir = self.join_dir('raw', create)
        self._archive_dir = self.join_dir('archive', create)
        self._frozen_file = self.join_file('.frozen')

        self._size = 0

    @classmethod
    def file_area(cls):
        return config.fs.staging

    @property
    def external_os_path(self):
        if not config.fs.staging_external:
            return self.os_path

        return self.os_path.replace(config.fs.staging, config.fs.staging_external)

    def to_staging_upload_files(self, create: bool = False, include_archive: bool = False) -> 'StagingUploadFiles':
        return self

    @property
    def size(self) -> int:
        return self._size

    def _file(self, path_object: PathObject, *args, **kwargs) -> IO:
        try:
            return open(path_object.os_path, *args, **kwargs)
        except FileNotFoundError:
            raise KeyError(path_object.os_path)
        except IsADirectoryError:
            raise KeyError(path_object.os_path)

    def is_empty(self) -> bool:
        return not os.path.exists(self._raw_dir.os_path) or not os.listdir(self._raw_dir.os_path)

    def raw_path_exists(self, path: str) -> bool:
        if not is_safe_relative_path(path):
            return False
        return os.path.exists(os.path.join(self._raw_dir.os_path, path))

    def raw_path_is_file(self, path: str) -> bool:
        if not is_safe_relative_path(path):
            return False
        return os.path.isfile(os.path.join(self._raw_dir.os_path, path))

    def raw_create_directory(self, path: str):
        assert path and is_safe_relative_path(path), 'Bad path provided'
        os.makedirs(os.path.join(self._raw_dir.os_path, path))

    def raw_directory_list(
            self, path: str = '', recursive=False, files_only=False, path_prefix=None) -> Iterable[RawPathInfo]:
        if not is_safe_relative_path(path):
            return
        os_path = os.path.join(self._raw_dir.os_path, path)
        if not os.path.isdir(os_path):
            return
        for element_name in sorted(os.listdir(os_path)):
            element_raw_path = os.path.join(path, element_name)
            element_os_path = os.path.join(os_path, element_name)
            is_file = os.path.isfile(element_os_path)
            if not is_file:
                # Crawl sub directory.
                dir_size = 0
                for sub_path_info in self.raw_directory_list(element_raw_path, True, files_only):
                    if sub_path_info.is_file:
                        dir_size += sub_path_info.size
                    if recursive:
                        if not path_prefix or sub_path_info.path.startswith(path_prefix):
                            yield sub_path_info

            if not files_only or is_file:
                size = os.stat(element_os_path).st_size if is_file else dir_size
                if not path_prefix or element_raw_path.startswith(path_prefix):
                    yield RawPathInfo(
                        path=element_raw_path,
                        is_file=is_file,
                        size=size,
                        access='unpublished')

    def raw_file(self, file_path: str, *args, **kwargs) -> IO:
        assert is_safe_relative_path(file_path)
        return self._file(self.raw_file_object(file_path), *args, **kwargs)

    def raw_file_size(self, file_path: str) -> int:
        assert is_safe_relative_path(file_path)
        return self.raw_file_object(file_path).size

    def raw_file_object(self, file_path: str) -> PathObject:
        assert is_safe_relative_path(file_path)
        return self._raw_dir.join_file(file_path)

    def write_archive(self, entry_id: str, data: Any) -> int:
        ''' Writes the data as archive file and returns the archive file size. '''
        archive_file_object = self.archive_file_object(entry_id)
        try:
            write_archive(archive_file_object.os_path, 1, data=[(entry_id, data)])
        except Exception as e:
            # in case of failure, remove the possible corrupted archive file
            if archive_file_object.exists():
                archive_file_object.delete()

            raise e

        return self.archive_file_object(entry_id).size

    def read_archive(self, entry_id: str) -> ArchiveReader:
        try:
            return read_archive(self.archive_file_object(entry_id).os_path)

        except FileNotFoundError:
            raise KeyError(entry_id)

    def archive_file_object(self, entry_id: str) -> PathObject:
        return self._archive_dir.join_file(f'{entry_id}.msg')

    def add_rawfiles(
            self, path: str, target_dir: str = '', cleanup_source_file_and_dir: bool = False) -> None:
        '''
        Adds the file or folder specified by `path` to this upload, in the raw directory
        specified by `target_dir`. If `path` denotes a zip or tar archive file, it will
        first be extracted to a temporary directory. The file(s) are *merged* with the
        existing upload files, i.e. new files are added, replacing old files if there
        already exists file(s) by the same names, the rest of the old files are left
        untouched.

        Cleanup
        The method is responsible for trying to clean up temporarily extracted files.
        If `cleanup_source_file_and_dir` is True, the source file (defined by `path`), and
        its parent directory (which we also assume is temporary) are also cleaned up.
        Note: the cleanup steps are always carried out, also if the operation fails.

        Arguments:
            path: OS path to a file or folder to add.
            target_dir: A raw path (i.e. path relative to the raw directory) defining
                where the resource defined by `path` should be put. If `target_dir` is not
                specified, it defaults to the empty string, i.e. the upload's raw dir.
            cleanup_source_file_and_dir: If true, the source file (defined by `path`) and
                its parent folder are included in the cleanup step - i.e. they are always
                deleted. Use when the file is stored temporarily.
        '''
        tmp_dir = None
        try:
            assert not self.is_frozen
            assert os.path.exists(path), f'{path} does not exist'
            assert is_safe_relative_path(target_dir)
            ext = os.path.splitext(path)[1]
            self._size += os.stat(path).st_size

            is_dir = os.path.isdir(path)
            if is_dir:
                is_zipfile = is_tarfile = False
            else:
                is_zipfile = zipfile.is_zipfile(path) or ext in zip_file_extensions
                is_tarfile = tarfile.is_tarfile(path) or ext in tar_file_extensions
                if is_zipfile or is_tarfile:
                    tmp_dir = create_tmp_dir(self.upload_id + '_unzip')
                    if is_zipfile:
                        with zipfile.ZipFile(path) as zf:
                            zf.extractall(tmp_dir)
                    elif is_tarfile:
                        with tarfile.open(path) as tf:
                            tf.extractall(tmp_dir)

            # Determine what to merge
            elements_to_merge: Iterable[Tuple[str, List[str], List[str]]] = []
            if is_dir or is_zipfile or is_tarfile:
                # Directory
                source_dir = path if is_dir else tmp_dir
                elements_to_merge = os.walk(source_dir)
            else:
                # Single file
                source_dir = os.path.dirname(path)
                elements_to_merge = [(source_dir, [], [os.path.basename(path)])]

            # Ensure target_dir exists and is a directory. If one of the elements in the
            # directory chain is a file, it needs to be deleted (the regular os.makedirs
            # doesn't do that).
            target_dir_subpath = self._raw_dir.os_path
            for dir_name in target_dir.split(os.path.sep):
                target_dir_subpath = os.path.join(target_dir_subpath, dir_name)
                if os.path.isfile(target_dir_subpath):
                    os.remove(target_dir_subpath)
                if not os.path.isdir(target_dir_subpath):
                    os.makedirs(target_dir_subpath)

            # Do the merge
            for root, dirs, files in elements_to_merge:
                elements = dirs + files
                os_target_dir = os.path.join(self._raw_dir.os_path, target_dir)
                for element in elements:
                    element_source_path = os.path.join(root, element)
                    element_relative_path = os.path.relpath(element_source_path, source_dir)
                    element_target_path = os.path.join(os_target_dir, element_relative_path)
                    if os.path.islink(element_source_path):
                        continue  # Skip links, could pose security risk
                    if os.path.exists(element_target_path):
                        if not (os.path.isdir(element_source_path) and os.path.isdir(element_target_path)):
                            # Target already exists and needs to be deleted
                            if os.path.isdir(element_target_path):
                                shutil.rmtree(element_target_path)
                            else:
                                os.remove(element_target_path)
                    # Copy or move the element
                    if os.path.isdir(element_source_path):
                        # Directory - just create corresponding directory in the target if needed.
                        if not os.path.exists(element_target_path):
                            os.makedirs(element_target_path)
                    else:
                        # File - copy or move it
                        if cleanup_source_file_and_dir or is_zipfile or is_tarfile:
                            # Move the file
                            shutil.move(element_source_path, element_target_path)
                        else:
                            # Copy the file
                            shutil.copyfile(element_source_path, element_target_path)
        finally:
            # Cleanup
            if tmp_dir and os.path.exists(tmp_dir):
                shutil.rmtree(tmp_dir)
            if cleanup_source_file_and_dir:
                if os.path.exists(path):
                    if os.path.isdir(path):
                        shutil.rmtree(path)
                    else:
                        os.remove(path)
                parent_dir = os.path.dirname(path)
                if os.path.exists(parent_dir):
                    shutil.rmtree(parent_dir)

    def delete_rawfiles(self, path):
        assert is_safe_relative_path(path)
        os_path = os.path.join(self.os_path, 'raw', path)
        assert os.path.exists(os_path)
        if os.path.isfile(os_path):
            os.remove(os_path)
        else:
            shutil.rmtree(os_path)
        if path == '':
            # Special case - deleting everything, i.e. the entire raw folder. Need to recreate.
            os.makedirs(os_path)

    @lru_cache()
    def metadata_file_cached(self, path_dir: str = ''):
        '''
        Gets the content of the metadata file located in the directory defined by `path_dir`.
        The `path_dir` should be relative to the `raw` folder.
        '''
        path_incl_filename = os.path.join(path_dir, config.process.metadata_file_name)
        for ext in config.process.metadata_file_extensions:
            path_incl_filename_ext = path_incl_filename + '.' + ext
            full_path = os.path.join(self._raw_dir.os_path, path_incl_filename_ext)
            if os.path.isfile(full_path):
                try:
                    with open(full_path) as f:
                        if full_path.endswith('.json'):
                            return json.load(f)
                        elif full_path.endswith('.yaml') or full_path.endswith('.yml'):
                            return yaml.load(f, Loader=getattr(yaml, 'FullLoader'))
                        else:
                            return {}
                except Exception as e:
                    # ignore the file contents if the file is not parsable, just warn.
                    self.logger.warn(
                        'could not parse nomad.yaml/json', path=path_incl_filename_ext, exc_info=e)
        return {}

    @property
    def is_frozen(self) -> bool:
        ''' Returns True if this upload is already *bagged*. '''
        return self._frozen_file.exists()

    def pack(
            self, entries: List[datamodel.EntryMetadata], with_embargo: bool, create: bool = True,
            include_raw: bool = True, include_archive: bool = True) -> None:
        '''
        Packs raw and/or archive files, to create the contents in the public file area.
        This method should be called when an upload is published, or when a
        published upload has been reprocessed.

        If the public upload files directory does not exist, it will be created.
        If the target archive file or raw file zip exists, they will be overwritten.
        If an archive file or raw file zip with the wrong access exists, they will be deleted.
        This is potentially a long running operation.

        Arguments:
            entries: A list of EntryMetadata to pack in the archive files
            with_embargo: If the upload is embargoed (determines which "access" is used in
                the file names)
            create: if the public upload files directory should be created. True by default.
            include_raw: determines if the raw data should be packed. True by default.
            include_archive: determines of the archive data should be packed. True by default.
        '''
        self.logger.info('started to pack upload')

        # freeze the upload
        assert not self.is_frozen, "Cannot pack an upload that is packed, or packing."
        with open(self._frozen_file.os_path, 'wt') as f:
            f.write('frozen')

        # Check embargo flag consistency
        for entry in entries:
            assert entry.with_embargo == with_embargo

        access = 'restricted' if with_embargo else 'public'
        other_access = 'public' if with_embargo else 'restricted'  # The "inverted" access

        # Get or create a target dir in the public area
        target_dir = DirectoryObject(PublicUploadFiles.base_folder_for(self.upload_id), create=create)
        if os.listdir(target_dir.os_path):
            # Target dir contains files. Check that the target access is identical
            assert PublicUploadFiles(self.upload_id).access == access, 'Inconsistent access'

        # zip archives
        if include_archive:
            with utils.timer(self.logger, 'packed msgpack archive') as log_data:
                number_of_entries = self._pack_archive_files(target_dir, entries, access, other_access)
                log_data.update(number_of_entries=number_of_entries)

        # zip raw files
        if include_raw:
            with utils.timer(self.logger, 'packed raw files'):
                self._pack_raw_files(target_dir, access, other_access)

    def _pack_archive_files(
            self, target_dir: DirectoryObject, entries: List[datamodel.EntryMetadata], access: str, other_access: str):
        number_of_entries = len(entries)

        def create_iterator():
            for entry in entries:
                archive_file = self.archive_file_object(entry.entry_id)
                if archive_file.exists():
                    data = read_archive(archive_file.os_path)[entry.entry_id].to_dict()
                    yield (entry.entry_id, data)
                else:
                    yield (entry.entry_id, {})

        try:
            file_object = PublicUploadFiles._create_msg_file_object(target_dir, access)
            write_archive(file_object.os_path, number_of_entries, create_iterator())
            # Remove the file with the opposite access, if it exists
            other_file_object = PublicUploadFiles._create_msg_file_object(target_dir, other_access)
            if other_file_object.exists():
                other_file_object.delete()  # This file should be empty, if it exists
        except Exception as e:
            self.logger.error('exception during packing archives', exc_info=e)
            raise

        return number_of_entries

    def _pack_raw_files(self, target_dir: DirectoryObject, access: str, other_access: str):
        try:
            raw_zip_file_object = PublicUploadFiles._create_raw_zip_file_object(target_dir, access)
            with zipfile.ZipFile(raw_zip_file_object.os_path, mode='w') as raw_zip:
                for path_info in self.raw_directory_list(recursive=True):
                    basename = os.path.basename(path_info.path)
                    if basename.startswith('POTCAR') and not basename.endswith('.stripped'):
                        continue  # Skip POTCAR files when publishing
                    raw_zip.write(self._raw_dir.join_file(path_info.path).os_path, path_info.path)
            # Remove the zip file with the opposite access, if it exists
            other_raw_zip_file_object = PublicUploadFiles._create_raw_zip_file_object(target_dir, other_access)
            if other_raw_zip_file_object.exists():
                other_raw_zip_file_object.delete()  # This file should be empty, if it exists
        except Exception as e:
            self.logger.error('exception during packing raw files', exc_info=e)
            raise

    def entry_files(self, mainfile: str, with_mainfile: bool = True, with_cutoff: bool = True) -> Iterable[str]:
        '''
        Returns all the auxfiles and mainfile for a given mainfile. This implements
        nomad's logic about what is part of an entry and what not. The mainfile
        is the first element, the rest is sorted.
        Arguments:
            mainfile: The mainfile path relative to upload
            with_mainfile: Do include the mainfile, default is True
        '''
        mainfile_object = self._raw_dir.join_file(mainfile)
        if not mainfile_object.exists():
            raise KeyError(mainfile)

        mainfile_basename = os.path.basename(mainfile)
        entry_dir = os.path.dirname(mainfile_object.os_path)
        entry_relative_dir = entry_dir[len(self._raw_dir.os_path) + 1:]

        file_count = 0
        aux_files: List[str] = []
        dir_elements = os.listdir(entry_dir)
        dir_elements.sort()
        for dir_element in dir_elements:
            if dir_element != mainfile_basename and os.path.isfile(os.path.join(entry_dir, dir_element)):
                aux_files.append(os.path.join(entry_relative_dir, dir_element))
                file_count += 1

            if with_cutoff and file_count > config.auxfile_cutoff:
                # If there are too many of them, its probably just a directory with lots of
                # mainfiles/entries. In this case it does not make any sense to provide thousands of
                # aux files.
                break

        aux_files = sorted(aux_files)

        if with_mainfile:
            return [mainfile] + aux_files
        else:
            return aux_files

    def entry_hash(self, mainfile: str, mainfile_key: str) -> str:
        '''
        Calculates a hash for the given entry based on file contents and aux file contents.
        Arguments:
            mainfile: The mainfile path relative to the upload that identifies the entry in
                the folder structure.
            mainfile_key: The mainfile_key of the entry (if any)
        Returns:
            The calculated hash
        Raises:
            KeyError: If the mainfile does not exist.
        '''
        hash = hashlib.sha512()
        for filepath in self.entry_files(mainfile):
            with open(self._raw_dir.join_file(filepath).os_path, 'rb') as f:
                for data in iter(lambda: f.read(65536), b''):
                    hash.update(data)
        if mainfile_key:
            hash.update(mainfile_key.encode('utf8'))
        return utils.make_websave(hash)

    def files_to_bundle(
            self, bundle_info: Dict[str, Any],
            include_raw_files: bool, include_archive_files: bool) -> FileSource:
        # Files to export for staging uploads.
        rv = CombinedFileSource()
        rv.add_file_source(StreamedFileSource(json_to_streamed_file(bundle_info, bundle_info_filename)))
        if include_raw_files:
            rv.add_file_source(DiskFileSource(self.os_path, 'raw'))
        if include_archive_files:
            rv.add_file_source(DiskFileSource(self.os_path, 'archive'))
        return rv

    @classmethod
    def files_from_bundle(
            cls, bundle_file_source: BrowsableFileSource,
            include_raw_files: bool,
            include_archive_files: bool,
            include_bundle_info: bool) -> FileSource:
        # Files to import for a staging upload
        rv = CombinedFileSource()
        if include_raw_files:
            rv.add_file_source(bundle_file_source.sub_source('raw'))
        if include_archive_files:
            rv.add_file_source(bundle_file_source.sub_source('archive'))
        if include_bundle_info:
            rv.add_file_source(bundle_file_source.sub_source(bundle_info_filename))
        return rv


class PublicUploadFiles(UploadFiles):

    def __init__(self, upload_id: str, create: bool = False):
        super().__init__(upload_id, create)
        self._directories: Dict[str, Dict[str, RawPathInfo]] = None
        self._raw_zip_file_object: PathObject = None
        self._raw_zip_file: zipfile.ZipFile = None
        self._archive_msg_file_object: PathObject = None
        self._archive_msg_file: ArchiveReader = None
        self._access: str = None

    @classmethod
    def file_area(cls):
        return config.fs.public

    @property
    def external_os_path(self):
        if not config.fs.public_external:
            return self.os_path

        self.os_path.replace(config.fs.public, config.fs.public_external)

    def close(self):
        if self._raw_zip_file is not None:
            self._raw_zip_file.close()

        if self._archive_msg_file is not None:
            self._archive_msg_file.close()

    @property
    def access(self):
        '''
        Which "access" is used, either 'public' (uploads without embargo) or 'restricted'
        (uploads with embargo). This is reflected in the names of the files holding the
        raw data and the archive data. The reason for this is so that it should be easy to
        see, by just looking at the files, if a published upload is embargoed or not.

        The access is determined by inspecting which files exist/contain data. If both
        public and restricted files exist/contain data, or if neither exists/contain data,
        a KeyError will be thrown (this should not happen if the upload is correctly packed).
        The inspection of the files is only done on the first call, and the cached result
        is used in subsequent calls. The only way to change the access is to call :func:`re_pack`.
        '''
        if self._access:
            return self._access
        # Determine access by inspecting the files
        files_found = False
        for access in ('public', 'restricted'):
            raw_zip_file_object = PublicUploadFiles._create_raw_zip_file_object(self, access)
            archive_msg_file_object = PublicUploadFiles._create_msg_file_object(self, access)
            found = (raw_zip_file_object.exists() and raw_zip_file_object.size > empty_zip_file_size) or (
                archive_msg_file_object.exists() and archive_msg_file_object.size > empty_archive_file_size)
            if found:
                if files_found:
                    self._access = self._raw_zip_file_object = self._archive_msg_file_object = None
                    raise KeyError('Inconsistency: both public and restricted files found')
                files_found = True
                self._raw_zip_file_object = raw_zip_file_object
                self._archive_msg_file_object = archive_msg_file_object
                self._access = access
        if not files_found:
            raise KeyError('Neither public nor restricted files found')
        return self._access

    @staticmethod
    def _create_raw_zip_file_object(target_dir: DirectoryObject, access: str) -> PathObject:
        return target_dir.join_file(f'raw-{access}.plain.zip')

    def raw_zip_file_object(self) -> PathObject:
        '''
        Gets the raw zip file, either public or restricted, depending on which one is used.
        If both public and restricted files exist, or if none of them exist, a KeyError will
        be thrown.
        '''
        self.access  # Invoke to initialize
        return self._raw_zip_file_object

    def _open_raw_zip_file(self) -> zipfile.ZipFile:
        if self._raw_zip_file:
            return self._raw_zip_file

        zip_path = self.raw_zip_file_object().os_path
        self._raw_zip_file = zipfile.ZipFile(zip_path)

        return self._raw_zip_file

    @staticmethod
    def _create_msg_file_object(target_dir: DirectoryObject, access: str) -> PathObject:
        version_suffix = '-' + config.fs.archive_version_suffix if config.fs.archive_version_suffix else ''
        return target_dir.join_file(f'archive-{access}{version_suffix}.msg.msg')

    def msg_file_object(self) -> PathObject:
        '''
        Gets the msg file, either public or restricted, depending on which one is used.
        If both public and restricted files exist, or if none of them exist, a KeyError will
        be thrown.
        '''
        self.access  # Invoke to initialize
        return self._archive_msg_file_object

    def _open_msg_file(self) -> ArchiveReader:
        if self._archive_msg_file is not None:
            if not self._archive_msg_file.is_closed():
                return self._archive_msg_file

        msg_file_object = self.msg_file_object()

        if not msg_file_object.exists():
            raise FileNotFoundError()

        archive = read_archive(msg_file_object.os_path)
        assert archive is not None
        self._archive_msg_file = archive

        return archive

    def to_staging_upload_files(self, create: bool = False, include_archive: bool = False) -> 'StagingUploadFiles':
        exists = StagingUploadFiles.exists_for(self.upload_id)
        if exists:
            if create:
                raise FileExistsError('Staging upload does already exist')
            return StagingUploadFiles(self.upload_id)
        elif not create:
            return None

        staging_upload_files = StagingUploadFiles(self.upload_id, create=True)
        # Extract files
        raw_zip_file = self.raw_zip_file_object()
        if raw_zip_file.exists():
            staging_upload_files.add_rawfiles(raw_zip_file.os_path)

        if include_archive:
            with self._open_msg_file() as archive:
                for entry_id, data in archive.items():
                    entry_id = entry_id.strip()
                    staging_upload_files.write_archive(entry_id, data.to_dict())

        return staging_upload_files

    def _parse_content(self):
        '''
        Parses the content of files and folders and caches it in self._directories for
        faster future access.
        '''
        if self._directories is None:
            self._directories = dict()
            self._directories[''] = {}  # Root folder
            directory_sizes: Dict[str, int] = {}
            # Add file RawPathInfo objects and calculate directory sizes
            try:
                zf = self._open_raw_zip_file()
                for path in zf.namelist():
                    file_name = os.path.basename(path)
                    directory_path = os.path.dirname(path)
                    size = zf.getinfo(path).file_size if file_name else 0

                    if directory_path:
                        # Ensure that all parent directories are added
                        sub_path = ''
                        for directory in directory_path.split(os.path.sep):
                            sub_path_next = os.path.join(sub_path, directory)
                            if sub_path_next not in self._directories:
                                self._directories[sub_path_next] = {}
                                directory_sizes.setdefault(sub_path_next, 0)
                            directory_sizes[sub_path_next] += size
                            sub_path = sub_path_next

                    if file_name:
                        directory_content = self._directories[directory_path]
                        directory_content[file_name] = RawPathInfo(
                            path=path,
                            is_file=True,
                            size=size,
                            access=self.access)
            except FileNotFoundError:
                pass
            # Add directories with the calculated sizes.
            for path, size in directory_sizes.items():
                basename = os.path.basename(path)
                directory_path = os.path.dirname(path)
                self._directories[directory_path][basename] = RawPathInfo(
                    path=path, is_file=False, size=size, access=self.access)

    def is_empty(self) -> bool:
        self._parse_content()
        return not self._directories.get('')

    def raw_path_exists(self, path: str) -> bool:
        if not is_safe_relative_path(path):
            return False
        self._parse_content()
        explicit_directory_path = path.endswith(os.path.sep)
        path = path.rstrip(os.path.sep)
        base_name = os.path.basename(path)
        directory_path = os.path.dirname(path)
        directory_content = self._directories.get(directory_path)
        if directory_content is not None:
            if not base_name:
                return True
            if base_name in directory_content:
                path_info = directory_content[base_name]
                if explicit_directory_path and path_info.is_file:
                    return False
                return True
        return False

    def raw_path_is_file(self, path: str) -> bool:
        if not is_safe_relative_path(path):
            return False
        self._parse_content()
        base_name = os.path.basename(path)
        directory_path = os.path.dirname(path)
        if not base_name:
            return False  # Requested path is an explicit directory path
        directory_content = self._directories.get(directory_path)
        if directory_content and base_name in directory_content:
            path_info = directory_content[base_name]
            return path_info.is_file
        return False

    def raw_directory_list(
            self, path: str = '', recursive=False, files_only=False, path_prefix=None) -> Iterable[RawPathInfo]:
        if not is_safe_relative_path(path):
            return
        self._parse_content()
        path = path.rstrip(os.path.sep)
        directory_content = self._directories.get(path)
        if directory_content is not None:
            for __, path_info in sorted(directory_content.items()):
                if not files_only or path_info.is_file:
                    if not path_prefix or path_info.path.startswith(path_prefix):
                        yield path_info
                if recursive and not path_info.is_file:
                    for sub_path_info in self.raw_directory_list(path_info.path, recursive, files_only):
                        if not path_prefix or sub_path_info.path.startswith(path_prefix):
                            yield sub_path_info

    def raw_file(self, file_path: str, *args, **kwargs) -> IO:
        assert is_safe_relative_path(file_path)
        mode = kwargs.get('mode') if len(args) == 0 else args[0]
        if 'mode' in kwargs:
            del(kwargs['mode'])
        mode = mode if mode else 'rb'

        try:
            zf = self._open_raw_zip_file()
            f = zf.open(file_path, 'r', **kwargs)
            if 't' in mode:
                return io.TextIOWrapper(f)
            else:
                return f
        except FileNotFoundError:
            pass
        except IsADirectoryError:
            pass
        except KeyError:
            pass

        raise KeyError(file_path)

    def raw_file_size(self, file_path: str) -> int:
        assert is_safe_relative_path(file_path)
        try:
            zf = self._open_raw_zip_file()
            info = zf.getinfo(file_path)
            return info.file_size
        except FileNotFoundError:
            pass
        except KeyError:
            pass

        raise KeyError(file_path)

    def read_archive(self, entry_id: str) -> Any:
        try:
            archive = self._open_msg_file()
            if entry_id in archive:
                return archive
        except FileNotFoundError:
            pass

        raise KeyError(entry_id)

    def re_pack(self, with_embargo: bool) -> None:
        '''
        Repacks the files when changing the embargo flag on the upload. That is: when lifting the
        embargo the file names of the raw zip file and the archive file change from containing
        the keyword "restricted" to "public". Adding embargo to a non-embargoed published
        upload is also supported, but only admins should be allowed to do this. The existing
        files are just renamed, so this should be a rather quick operation. The upload must
        be correctly packed (i.e. there cannot be non-empty public and restricted files
        at the same time).
        '''
        if (self.access == 'restricted' and with_embargo) or (self.access == 'public' and not with_embargo):
            return  # Nothing to do
        self.close()
        new_access = 'restricted' if with_embargo else 'public'
        msg_file_object = self.msg_file_object()
        msg_file_object_new = PublicUploadFiles._create_msg_file_object(self, new_access)
        if msg_file_object.exists():
            if msg_file_object_new.exists():
                msg_file_object_new.delete()  # We have checked that the file is empty anyway
            os.rename(msg_file_object.os_path, msg_file_object_new.os_path)
        raw_zip_file_object = self.raw_zip_file_object()
        raw_zip_file_object_new = PublicUploadFiles._create_raw_zip_file_object(self, new_access)
        if raw_zip_file_object.exists():
            if raw_zip_file_object_new.exists():
                raw_zip_file_object_new.delete()  # We have checked that the file is empty anyway
            os.rename(raw_zip_file_object.os_path, raw_zip_file_object_new.os_path)

        # Clear the cached values
        self._access = None
        self._raw_zip_file = self._raw_zip_file_object = None
        self._archive_msg_file = self._archive_msg_file_object = None

    def files_to_bundle(
            self, bundle_info: Dict[str, Any],
            include_raw_files: bool, include_archive_files: bool) -> FileSource:
        # Defines files for upload bundles of published uploads.
        rv = CombinedFileSource()
        rv.add_file_source(StreamedFileSource(json_to_streamed_file(bundle_info, bundle_info_filename)))
        for filename in os.listdir(self.os_path):
            if filename.startswith('raw-') and include_raw_files:
                rv.add_file_source(DiskFileSource(self.os_path, filename))
            if filename.startswith('archive-') and include_archive_files:
                rv.add_file_source(DiskFileSource(self.os_path, filename))
        return rv

    @classmethod
    def files_from_bundle(
            cls, bundle_file_source: BrowsableFileSource,
            include_raw_files: bool,
            include_archive_files: bool,
            include_bundle_info: bool) -> FileSource:
        rv = CombinedFileSource()
        for filename in bundle_file_source.directory_list(''):
            if filename.startswith('raw-') and include_raw_files:
                rv.add_file_source(bundle_file_source.sub_source(filename))
            if filename.startswith('archive-') and include_archive_files:
                rv.add_file_source(bundle_file_source.sub_source(filename))
            if filename == bundle_info_filename and include_bundle_info:
                rv.add_file_source(bundle_file_source.sub_source(filename))
        return rv


class UploadBundle:
    '''
    Class for handling file-related logic for an *upload bundle*. Upload bundles are used
    to import and export uploads between different NOMAD installations.
    '''
    def __init__(self, path: str):
        ''' Creates an UploadBundle instance. The `path` should denote a zipfile or a folder. '''
        self.path = path
        self.file_source: BrowsableFileSource = None
        self._bundle_info: Dict[str, Any] = None
        if os.path.isdir(path):
            self.file_source = DiskFileSource(path)
        else:
            assert zipfile.is_zipfile(path), '`path` must define a folder or a zipfile.'
            zip_file = zipfile.ZipFile(path, 'r')
            self.file_source = ZipFileSource(zip_file)

    @property
    def bundle_info(self) -> Dict[str, Any]:
        if self._bundle_info is None:
            with self.file_source.open(bundle_info_filename, 'rt') as f:
                self._bundle_info = json.load(f, cls=StandardJSONDecoder)
        return self._bundle_info

    def import_upload_files(
            self, include_raw_files: bool, include_archive_files: bool, include_bundle_info: bool,
            move_files: bool) -> UploadFiles:
        '''
        Creates an :class:`UploadFiles` object of the right type and imports the selected
        files to it. The target folder must not already exist.
        '''
        try:
            upload_files: UploadFiles = None
            upload_id: str = self.bundle_info['upload_id']
            published: bool = self.bundle_info.get('upload', {}).get('publish_time') is not None
            cls = PublicUploadFiles if published else StagingUploadFiles
            assert not os.path.exists(cls.base_folder_for(upload_id)), 'Upload folder already exists'
            upload_files = cls(upload_id, create=True)
            import_file_source = upload_files.files_from_bundle(
                self.file_source, include_raw_files, include_archive_files, include_bundle_info)
            import_file_source.to_disk(upload_files.os_path, move_files=move_files, overwrite=True)
            return upload_files
        except Exception:
            # Some thing went wrong. Delete the files and re-raise the original exception
            if upload_files:
                upload_files.delete()
            raise

    def close(self):
        self.file_source.close()

    def delete(self, include_parent_folder: bool = False):
        '''
        Deletes the bundle files. If `include_parent_folder` is set, and the parent folder
        is empty, it is also deleted.
        '''
        self.close()
        if os.path.exists(self.path):
            PathObject(self.path).delete()
        if include_parent_folder:
            parent_folder = os.path.dirname(self.path)
            if not os.listdir(parent_folder):
                PathObject(parent_folder).delete()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import datetime
import numpy as np
import ase.io
from os import path

from nomad.datamodel import EntryArchive
from nomad.units import ureg as units
from nomad.datamodel.metainfo.simulation.run import Run, Program, TimeRun
from nomad.datamodel.metainfo.simulation.system import (
    System, Atoms)
from nomad.datamodel.metainfo.simulation.method import (
    Method, Electronic, BasisSet)
from nomad.datamodel.metainfo.simulation.calculation import (
    Calculation, Dos, DosValues, Charges)

from nomad.parsing.file_parser import TextParser, Quantity

from .metainfo.lobster import x_lobster_section_cohp, x_lobster_section_coop

'''
This is a LOBSTER code parser.
'''

e = (1 * units.e).to_base_units().magnitude
eV = (1 * units.eV).to_base_units().magnitude


def parse_ICOXPLIST(fname, scc, method):

    def icoxp_line_split(string):
        tmp = string.split()
        # LOBSTER version 3 and above
        if len(tmp) == 8:
            return [tmp[1], tmp[2], float(tmp[3]), [int(tmp[4]),
                    int(tmp[5]), int(tmp[6])], float(tmp[7])]
        # LOBSTER versions below 3
        elif len(tmp) == 6:
            return [tmp[1], tmp[2], float(tmp[3]), float(tmp[4]), int(tmp[5])]

    icoxplist_parser = TextParser(quantities=[
        Quantity('icoxpslist_for_spin', r'\s*CO[OH]P.*spin\s*\d\s*([^#]+[-\d\.]+)',
                 repeats=True,
                 sub_parser=TextParser(quantities=[
                     Quantity('line',
                              # LOBSTER version 3 and above
                              r'(\s*\d+\s+\w+\s+\w+\s+[\.\d]+\s+[-\d]+\s+[-\d]+\s+[-\d]+\s+[-\.\d]+\s*)|'
                              # LOBSTER versions below 3
                              r'(\s*\d+\s+\w+\s+\w+\s+[\.\d]+\s+[-\.\d]+\s+[\d]+\s*)',
                              repeats=True, str_operation=icoxp_line_split)])
                 )
    ])

    if not path.isfile(fname):
        return
    icoxplist_parser.mainfile = fname
    icoxplist_parser.parse()

    icoxp = []
    for spin, icoxplist in enumerate(icoxplist_parser.get('icoxpslist_for_spin')):

        lines = icoxplist.get('line')
        if lines is None:
            break
        if type(lines[0][4]) is int:
            a1, a2, distances, tmp, bonds = zip(*lines)
        else:
            a1, a2, distances, v, tmp = zip(*lines)
        icoxp.append(0)
        icoxp[-1] = list(tmp)
        if spin == 0:
            if method == 'o':
                section = scc.m_create(x_lobster_section_coop)
            elif method == 'h':
                section = scc.m_create(x_lobster_section_cohp)

            setattr(section, "x_lobster_number_of_co{}p_pairs".format(
                method), len(list(a1)))
            setattr(section, "x_lobster_co{}p_atom1_labels".format(
                method), list(a1))
            setattr(section, "x_lobster_co{}p_atom2_labels".format(
                method), list(a2))
            setattr(section, "x_lobster_co{}p_distances".format(
                method), np.array(distances) * units.angstrom)

            # version specific entries
            if 'v' in locals():
                setattr(section, "x_lobster_co{}p_translations".format(
                    method), list(v))
            if 'bonds' in locals():
                setattr(section, "x_lobster_co{}p_number_of_bonds".format(
                    method), list(bonds))

    if len(icoxp) > 0:
        setattr(section, "x_lobster_integrated_co{}p_at_fermi_level".format(
            method), np.array(icoxp) * units.eV)


def parse_COXPCAR(fname, scc, method, logger):
    coxpcar_parser = TextParser(quantities=[
        Quantity('coxp_pairs', r'No\.\d+:(\w{1,2}\d+)->(\w{1,2}\d+)\(([\d\.]+)\)\s*?',
                 repeats=True),
        Quantity('coxp_lines', r'\n\s*(-*\d+\.\d+(?:[ \t]+-*\d+\.\d+)+)',
                 repeats=True)
    ])

    if not path.isfile(fname):
        return
    coxpcar_parser.mainfile = fname
    coxpcar_parser.parse()

    if method == 'o':
        if not scc.x_lobster_section_coop:
            section = scc.m_create(x_lobster_section_coop)
        else:
            section = scc.x_lobster_section_coop
    elif method == 'h':
        if not scc.x_lobster_section_cohp:
            section = scc.m_create(x_lobster_section_cohp)
        else:
            section = scc.x_lobster_section_cohp

    pairs = coxpcar_parser.get('coxp_pairs')
    if pairs is None:
        logger.warning('No CO{}P values detected in CO{}PCAR.lobster.'.format(
            method.upper(), method.upper()))
        return
    a1, a2, distances = zip(*pairs)
    number_of_pairs = len(list(a1))

    setattr(section, "x_lobster_number_of_co{}p_pairs".format(
        method), number_of_pairs)
    setattr(section, "x_lobster_co{}p_atom1_labels".format(
        method), list(a1))
    setattr(section, "x_lobster_co{}p_atom2_labels".format(
        method), list(a2))
    setattr(section, "x_lobster_co{}p_distances".format(
        method), np.array(distances) * units.angstrom)

    coxp_lines = coxpcar_parser.get('coxp_lines')
    if coxp_lines is None:
        logger.warning('No CO{}P values detected in CO{}PCAR.lobster.'
                       'The file is likely incomplete'.format(
                           method.upper(), method.upper()))
        return
    coxp_lines = list(zip(*coxp_lines))

    setattr(section, "x_lobster_number_of_co{}p_values".format(
        method), len(coxp_lines[0]))
    setattr(section, "x_lobster_co{}p_energies".format(
        method), np.array(coxp_lines[0]) * units.eV)

    if len(coxp_lines) == 2 * number_of_pairs + 3:
        coxp = [[x] for x in coxp_lines[3::2]]
        icoxp = [[x] for x in coxp_lines[4::2]]
        acoxp = [coxp_lines[1]]
        aicoxp = [coxp_lines[2]]
    elif len(coxp_lines) == 4 * number_of_pairs + 5:
        coxp = [x for x in zip(coxp_lines[5:number_of_pairs * 2 + 4:2],
                coxp_lines[number_of_pairs * 2 + 5: 4 * number_of_pairs + 4:2])]
        icoxp = [x for x in zip(coxp_lines[6:number_of_pairs * 2 + 5:2],
                 coxp_lines[number_of_pairs * 2 + 6: 4 * number_of_pairs + 5:2])]
        acoxp = [coxp_lines[1], coxp_lines[3]]
        aicoxp = [coxp_lines[2], coxp_lines[4]]
    else:
        logger.warning('Unexpected number of columns {} '
                       'in CO{}PCAR.lobster.'.format(len(coxp_lines),
                                                     method.upper()))
        return

    # FIXME: correct magnitude?
    setattr(section, "x_lobster_co{}p_values".format(
        method), np.array(coxp))
    setattr(section, "x_lobster_average_co{}p_values".format(
        method), np.array(acoxp))
    setattr(section, "x_lobster_integrated_co{}p_values".format(
        method), np.array(icoxp) * units.eV)
    setattr(section, "x_lobster_average_integrated_co{}p_values".format(
        method), np.array(aicoxp) * units.eV)
    setattr(section, "x_lobster_integrated_co{}p_values".format(
        method), np.array(icoxp) * units.eV)


def parse_CHARGE(fname, scc):
    charge_parser = TextParser(quantities=[
        Quantity(
            'charges', r'\s*\d+\s+[A-Za-z]{1,2}\s+([-\d\.]+)\s+([-\d\.]+)\s*', repeats=True)
    ])

    if not path.isfile(fname):
        return
    charge_parser.mainfile = fname
    charge_parser.parse()

    charges = charge_parser.get('charges')
    if charges is not None:
        sec_charges = scc.m_create(Charges)
        sec_charges.analysis_method = "mulliken"
        sec_charges.kind = "integrated"
        sec_charges.value = np.array(list(zip(*charges))[0]) * units.elementary_charge
        sec_charges = scc.m_create(Charges)
        sec_charges.analysis_method = "loewdin"
        sec_charges.kind = "integrated"
        sec_charges.value = np.array(list(zip(*charges))[1]) * units.elementary_charge


def parse_DOSCAR(fname, run, logger):

    def parse_species(run, atomic_numbers):
        """
        If we don't have any structure from the underlying DFT code, we can
        at least figure out what atoms we have in the structure. The best place
        to get this info from is the DOSCAR.lobster
        """

        if not run.system:
            system = run.m_create(System)
            system.atoms = Atoms(species=atomic_numbers, periodic=[True, True, True])

    def translate_lm(lm):
        lm_dictionary = {
            's': [0, 0],
            'p_z': [1, 0],
            'p_x': [1, 1],
            'p_y': [1, 2],
            'd_z^2': [2, 0],
            'd_xz': [2, 1],
            'd_yz': [2, 2],
            'd_xy': [2, 3],
            'd_x^2-y^2': [2, 4],
            'z^3': [3, 0],
            'xz^2': [3, 1],
            'yz^2': [3, 2],
            'xyz': [3, 3],
            'z(x^2-y^2)': [3, 4],
            'x(x^2-3y^2)': [3, 5],
            'y(3x^2-y^2)': [3, 6],
        }
        return lm_dictionary.get(lm[1:])

    if not path.isfile(fname):
        return

    energies = []
    dos_values = []
    integral_dos = []
    atom_projected_dos_values = []
    atom_index = 0
    n_atoms = 0
    n_dos = 0
    atomic_numbers = []
    lms = []
    with open(fname) as f:
        for i, line in enumerate(f):
            if i == 0:
                n_atoms = int(line.split()[0])
            if i == 1:
                _ = float(line.split()[0]) * units.angstrom**3
            if i == 5:
                n_dos = int(line.split()[2])
            if 'Z=' in line:
                atom_index += 1
                atom_projected_dos_values.append([])
                lms.append((line.split(';')[-1]).split())
                atomic_numbers.append(int(line.split(';')[-2].split('=')[1]))
                continue
            if i > 5:
                line = [float(x) for x in line.split()]
                if atom_index == 0:
                    energies.append(line[0])
                    if len(line) == 3:
                        dos_values.append([line[1]])
                        integral_dos.append([line[2]])
                    elif len(line) == 5:
                        dos_values.append([line[1], line[2]])
                        integral_dos.append([line[3], line[4]])
                else:
                    atom_projected_dos_values[-1].append(line[1:])

    if len(atomic_numbers) > 0 and len(atomic_numbers) == n_atoms:
        parse_species(run, atomic_numbers)

    if n_dos == 0:
        return

    if len(dos_values) == n_dos:
        dos = run.calculation[0].m_create(Dos, Calculation.dos_electronic)
        dos.n_energies = n_dos
        dos.energies = energies * units.eV
        value = list(zip(*dos_values))
        n_electrons = sum(atomic_numbers)
        index = (np.abs(energies)).argmin()
        # integrated dos at the Fermi level should be the number of electrons
        n_valence_electrons = int(round(sum(integral_dos[index])))
        n_core_electrons = n_electrons - n_valence_electrons
        value_integrated = np.array(list(zip(*integral_dos))) + n_core_electrons / len(integral_dos[0])
        for spin_i in range(len(value)):
            dos_total = dos.m_create(DosValues, Dos.total)
            dos_total.spin = spin_i
            dos_total.value = value[spin_i] * (1 / units.eV)
            dos_total.value_integrated = value_integrated[spin_i]
    else:
        logger.warning('Unable to parse total dos from DOSCAR.lobster, \
                            it doesn\'t contain enough dos values')
        return

    for atom_i, pdos in enumerate(atom_projected_dos_values):
        if len(pdos) != n_dos:
            logger.warning('Unable to parse atom lm-projected dos from DOSCAR.lobster, \
                            it doesn\'t contain enough dos values')
            continue

        if len(lms[atom_i]) == len(pdos[0]):
            # we have the same lm-projections for spin up and dn
            dos_values = np.array([[lmdos] for lmdos in zip(*pdos)]) / eV
        elif len(lms[atom_i]) * 2 == len(pdos[0]):
            pdos_up = list(zip(*pdos))[0::2]
            pdos_dn = list(zip(*pdos))[1::2]
            dos_values = np.array([[a, b] for a, b in zip(pdos_up, pdos_dn)]) / eV
        else:
            logger.warning('Unexpected number of columns in DOSCAR.lobster')
            return
        for lm_i, lm in enumerate(lms[atom_i]):
            for spin_i in range(len(dos_values[lm_i])):
                section_pdos = dos.m_create(DosValues, Dos.atom_projected)
                section_pdos.atom_index = atom_i
                section_pdos.spin = spin_i
                section_pdos.m_kind = 'real_orbital'
                section_pdos.lm = translate_lm(lm)
                section_pdos.value = dos_values[lm_i][spin_i]


mainfile_parser = TextParser(quantities=[
    Quantity('program_version', r'^LOBSTER\s*v([\d\.]+)\s*', repeats=False),
    Quantity('datetime', r'starting on host \S* on (\d{4}-\d\d-\d\d\sat\s\d\d:\d\d:\d\d)\s[A-Z]{3,4}',
             repeats=False),
    Quantity('x_lobster_code',
             r'detecting used PAW program... (.*)', repeats=False),
    Quantity('x_lobster_basis',
             r'setting up local basis functions...\s*((?:[a-zA-Z]{1,2}\s+\(.+\)(?:\s+\d\S+)+\s+)+)',
             repeats=False,
             sub_parser=TextParser(quantities=[
                 Quantity('x_lobster_basis_species',
                          r'([a-zA-Z]+){1,2}\s+\((.+)\)((?:\s+\d\S+)+)\s+', repeats=True)
             ])),
    Quantity('spilling', r'((?:spillings|abs. tot)[\s\S]*?charge\s*spilling:\s*\d+\.\d+%)',
             repeats=True,
             sub_parser=TextParser(quantities=[
                 Quantity('abs_total_spilling',
                          r'abs.\s*total\s*spilling:\s*(\d+\.\d+)%', repeats=False),
                 Quantity('abs_charge_spilling',
                          r'abs.\s*charge\s*spilling:\s*(\d+\.\d+)%', repeats=False)
             ])),
    Quantity('finished', r'finished in (\d)', repeats=False),
])


class LobsterParser:
    def __init__(self):
        pass

    def parse(self, mainfile: str, archive: EntryArchive, logger=None):
        mainfile_parser.mainfile = mainfile
        mainfile_path = path.dirname(mainfile)
        mainfile_parser.parse()

        run = archive.m_create(Run)

        run.program = Program(
            name='LOBSTER',
            version=str(mainfile_parser.get('program_version')))
        # FIXME: There is a timezone info present as well, but datetime support for timezones
        # is bad and it doesn't support some timezones (for example CEST).
        # That leads to test failures, so ignore it for now.
        date = datetime.datetime.strptime(' '.join(mainfile_parser.get('datetime')),
                                          '%Y-%m-%d at %H:%M:%S') - datetime.datetime(1970, 1, 1)
        run.time_run = TimeRun(wall_start=date.total_seconds())
        code = mainfile_parser.get('x_lobster_code')

        # parse structure
        if code is not None:
            if code == 'VASP':
                try:
                    structure = ase.io.read(mainfile_path + '/CONTCAR', format="vasp")
                except FileNotFoundError:
                    logger.warning('Unable to parse structure info, no CONTCAR detected')
            else:
                logger.warning('Parsing of {} structure is not supported'.format(code))
        if 'structure' in locals():
            system = run.m_create(System)
            system.atoms = Atoms(
                lattice_vectors=structure.get_cell() * units.angstrom,
                labels=structure.get_chemical_symbols(),
                periodic=structure.get_pbc(),
                positions=structure.get_positions() * units.angstrom)

        if mainfile_parser.get('finished') is not None:
            run.clean_end = True
        else:
            run.clean_end = False

        scc = run.m_create(Calculation)
        method = run.m_create(Method)
        scc.method_ref = method

        spilling = mainfile_parser.get('spilling')
        if spilling is not None:
            method.electronic = Electronic(n_spin_channels=len(spilling))
            total_spilling = []
            charge_spilling = []
            for s in spilling:
                total_spilling.append(s.get('abs_total_spilling'))
                charge_spilling.append(s.get('abs_charge_spilling'))
            scc.x_lobster_abs_total_spilling = np.array(total_spilling)
            scc.x_lobster_abs_charge_spilling = np.array(charge_spilling)

        method_keys = [
            'x_lobster_code'
        ]

        for key in method_keys:
            val = mainfile_parser.get(key)
            if val is not None:
                setattr(method, key, val)

        basis = mainfile_parser.get('x_lobster_basis')
        if basis is not None:
            species = basis.get('x_lobster_basis_species')
            if species is not None:
                method.basis_set.append(BasisSet(name=species[0][1]))

        parse_ICOXPLIST(mainfile_path + '/ICOHPLIST.lobster', scc, 'h')
        parse_ICOXPLIST(mainfile_path + '/ICOOPLIST.lobster', scc, 'o')

        parse_COXPCAR(mainfile_path + '/COHPCAR.lobster', scc, 'h', logger)
        parse_COXPCAR(mainfile_path + '/COOPCAR.lobster', scc, 'o', logger)

        parse_CHARGE(mainfile_path + '/CHARGE.lobster', scc)

        parse_DOSCAR(mainfile_path + '/DOSCAR.lobster', run, logger)

        if run.system:
            scc.system_ref = run.system[0]



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Union, Dict, List
from nptyping import NDArray
import re
import ase.data
from ase import Atoms
import numpy as np
import matid.geometry

from nomad.datamodel.results import Symmetry, Material
from nomad import atomutils
from nomad.utils import hash


class MaterialNormalizer():
    def __init__(
            self,
            entry_archive,
            repr_system,
            repr_symmetry,
            spg_number,
            conv_atoms,
            wyckoff_sets,
            optimade,
            logger):
        self.entry_archive = entry_archive
        self.repr_system = repr_system
        self.repr_method = None
        self.spg_number = spg_number
        self.conv_atoms = conv_atoms
        self.wyckoff_sets = wyckoff_sets
        self.optimade = optimade
        self.repr_symmetry = repr_symmetry
        self.structural_type = None
        self.run = entry_archive.run[0]
        self.logger = logger

    def material(self) -> Material:
        """Returns a populated Material subsection."""
        material = Material()

        if self.repr_system:
            self.structural_type = self.repr_system.type
            material.structural_type = self.repr_system.type
            symbols, reduced_counts = atomutils.get_hill_decomposition(self.repr_system.atoms.labels, reduced=True)
            material.chemical_formula_reduced_fragments = [
                "{}{}".format(n, int(c) if c != 1 else "") for n, c in zip(symbols, reduced_counts)
            ]
            material.chemical_formula_hill = self.repr_system.chemical_composition_hill
            material.chemical_formula_descriptive = material.chemical_formula_hill
        if self.optimade:
            material.elements = self.optimade.elements
            material.chemical_formula_reduced = self.optimade.chemical_formula_reduced
            material.chemical_formula_anonymous = self.optimade.chemical_formula_anonymous

        material.symmetry = self.symmetry()

        if self.structural_type == "bulk":
            material.material_id = self.material_id_bulk(self.spg_number, self.wyckoff_sets)
            material.material_name = self.material_name(symbols, reduced_counts)
            classes = self.material_classification()
            if classes:
                material.functional_type = classes.get("material_class_springer")
                material.compound_type = classes.get("compound_class_springer")
        if self.structural_type == "2D":
            material.material_id = self.material_id_2d(self.spg_number, self.wyckoff_sets)
        elif self.structural_type == "1D":
            material.material_id = self.material_id_1d(self.conv_atoms)

        return material

    def material_id_bulk(self, spg_number: int, wyckoff_sets) -> str:
        if spg_number is None or wyckoff_sets is None:
            return None
        norm_hash_string = atomutils.get_symmetry_string(spg_number, wyckoff_sets)
        return hash(norm_hash_string)

    def material_id_2d(self, spg_number: int, wyckoff_sets) -> str:
        if spg_number is None or wyckoff_sets is None:
            return None
        norm_hash_string = atomutils.get_symmetry_string(spg_number, wyckoff_sets, is_2d=True)
        return hash(norm_hash_string)

    def material_id_1d(self, conv_atoms: Atoms) -> str:
        """Hash to be used as identifier for a 1D material. Based on Coulomb
        matrix eigenvalues and the Hill formula.

        The fingerprint is based on calculating a discretized version of a
        sorted Coulomb matrix eigenspectrum (Grégoire Montavon, Katja Hansen,
        Siamac Fazli, Matthias Rupp, Franziska Biegler, Andreas Ziehe,
        Alexandre Tkatchenko, Anatole V. Lilienfeld, and Klaus-Robert Müller.
        Learning invariant representations of molecules for atomization energy
        prediction. In F. Pereira, C. J. C. Burges, L. Bottou, and K. Q.
        Weinberger, editors, Advances in Neural Information Processing Systems
        25, pages 440–448. Curran Associates, Inc., 2012.).

        The fingerprints are discretized in order to perform O(n) matching
        between structures (no need to compare fingerprints against each
        other). As regular discretization is susceptible to the "edge problem",
        a robust discretization is used instead (Birget, Jean-Camille & Hong,
        Dawei & Memon, Nasir. (2003). Robust discretization, with an
        application to graphical passwords. IACR Cryptology ePrint Archive.
        2003. 168.) Basically for the 1-dimensional domain two grids are
        created and the points are mapped to the first grid in which they are
        robust using a minimum tolerance parameter r, with the maximum
        tolerance being 5r.

        There are other robust discretization methods that can guarantee exact
        r-tolerance (e.g. Sonia Chiasson, Jayakumar Srinivasan, Robert Biddle,
        and P. C. van Oorschot. 2008. Centered discretization with application
        to graphical passwords. In Proceedings of the 1st Conference on
        Usability, Psychology, and Security (UPSEC’08). USENIX Association,
        USA, Article 6, 1–9.). This method however requires that a predefined
        "correct" structure exists against which the search is done.
        """
        if conv_atoms is None:
            return None

        # Calculate charge part
        q = conv_atoms.get_atomic_numbers()
        qiqj = np.sqrt(q[None, :] * q[:, None])

        # Calculate distance part. Notice that the minimum image convention
        # must be used. Without it, differently oriented atoms in the same cell
        # may be detected as the same material.
        pos = conv_atoms.get_positions()
        cell = conv_atoms.get_cell()
        cmat = 10 - matid.geometry.get_distance_matrix(pos, pos, cell, pbc=True, mic=True)
        cmat = np.clip(cmat, a_min=0, a_max=None)
        np.fill_diagonal(cmat, 0)
        cmat = qiqj * cmat

        # Calculate eigenvalues
        eigval, _ = np.linalg.eigh(cmat)

        # Sort eigenvalues
        eigval = np.array(sorted(eigval))

        # Perform robust discretization (see function docstring for details). r
        # = 0.5 ensures that all grids are integers which can be uniquely
        # mapped to strings. If finer grid is needed adjust the eigenvalue scale
        # instead.
        eigval /= 25  # Go to smaller scale where integer numbers are meaningful
        dimension = 1
        r = 0.5
        spacing = 2 * r * (dimension + 1)
        phi_k = 2 * r * np.array(range(dimension + 1))
        t = np.mod((eigval[None, :] + phi_k[:, None]), (2 * r * (dimension + 1)))
        grid_mask = (r <= t) & (t < r * (2 * dimension + 1))
        safe_grid_k = np.argmax(grid_mask == True, axis=0)   # noqa: E712
        discretization = spacing * np.floor((eigval + (2 * r * safe_grid_k)) / spacing)
        discretization[safe_grid_k == 1] += 2 * r

        # Construct formula
        names, counts = atomutils.get_hill_decomposition(conv_atoms.get_chemical_symbols(), reduced=False)
        formula = atomutils.get_formula_string(names, counts)

        # Form hash
        strings = []
        for number in discretization:
            num_str = str(int(number))
            strings.append(num_str)
        fingerprint = ";".join(strings)
        id_strings = []
        id_strings.append(formula)
        id_strings.append(fingerprint)
        hash_seed = ", ".join(id_strings)
        hash_val = hash(hash_seed)

        return hash_val

    def material_classification(self) -> Dict[str, List[str]]:
        try:
            sec_springer = self.repr_system["springer_material"][0]
        except Exception:
            return None

        classes: Dict[str, List[str]] = {}
        try:
            classifications = sec_springer['classification']
        except KeyError:
            pass
        else:
            classes["material_class_springer"] = classifications
        try:
            compound_classes = sec_springer['compound_class']
        except KeyError:
            pass
        else:
            classes["compound_class_springer"] = compound_classes
        return classes

    def material_name(
            self, symbols: Union[List, NDArray], counts: Union[List, NDArray]) -> str:
        if symbols is None or counts is None:
            return None
        name = None
        # Systems with one element are named after it
        if len(symbols) == 1:
            number = ase.data.atomic_numbers[symbols[0]]
            name = ase.data.atomic_names[number]
        # Binary systems have specific names
        elif len(symbols) == 2:
            atomicnumbers = [ase.data.atomic_numbers[i] for i in symbols]
            names = [ase.data.atomic_names[i] for i in atomicnumbers]

            # Non-metal elements are anions in the binary compounds and receive the -ide suffix
            if names[1] == "Antimony":
                names[1] = names[1][:-1] + "ide"
            if names[1] == "Arsenic":
                names[1] = names[1][:-1] + "de"
            if names[1] == "Boron" or names[1] == "Carbon":
                names[1] = names[1][:-2] + "ide"
            if names[1] == "Chlorine" or names[1] == "Germanium" or names[1] == "Selenium" or names[1] == "Bromine" \
               or names[1] == "Tellurium" or names[1] == "Iodine" or names[1] == "Polonium" or names[1] == "Astatine" or \
               names[1] == "Fluorine":
                names[1] = names[1][:-2] + "de"
            if names[1] == "Silicon" or names[1] == "Sulfur":
                names[1] = names[1][:-2] + "ide"
            if names[1] == "Nitrogen" or names[1] == "Oxygen" or names[1] == "Hydrogen" or names[1] == "Phosphorus":
                names[1] = names[1][:-4] + "ide"

            name = names[0] + " " + names[1]

            if names[1] == "Fluoride" or names[1] == "Chloride" or names[1] == "Bromide" or \
               names[1] == "Iodide" or names[1] == "Hydride":

                # Non-metals with elements of variable valence, therefore we remove alkaline and
                # alkaline-earth elements, which have fixed valence
                # Only the most electronegative non-metals are supposed to make ionic compounds
                if names[0] != "Lithium" and names[0] != "Sodium" and names[0] != "Potassium" and \
                   names[0] != "Rubidium" and names[0] != "Cesium" and names[0] != "Francium" and \
                   names[0] != "Beryllium" and names[0] != "Magnesium" and names[0] != "Calcium" and \
                   names[0] != "Strontium" and names[0] != "Barium" and names[0] != "Radium" and \
                   names[0] != "Aluminum":

                    if counts[1] == 2:
                        name = names[0] + "(II)" + " " + names[1]
                    elif counts[1] == 3:
                        name = names[0] + "(III)" + " " + names[1]
                    elif counts[1] == 4:
                        name = names[0] + "(IV)" + " " + names[1]
                    elif counts[1] == 5:
                        name = names[0] + "(V)" + " " + names[1]
                    elif counts[1] == 6:
                        name = names[0] + "(VI)" + " " + names[1]
                    elif counts[1] == 7:
                        name = names[0] + "(VII)" + " " + names[1]

            if names[1] == "Oxide" or names[1] == "Sulfide" or names[1] == "Selenide":
                if names[0] != "Lithium" and names[0] != "Sodium" and names[0] != "Potassium" and \
                   names[0] != "Rubidium" and names[0] != "Cesium" and names[0] != "Francium" and \
                   names[0] != "Beryllium" and names[0] != "Magnesium" and names[0] != "Calcium" and \
                   names[0] != "Strontium" and names[0] != "Barium" and names[0] != "Radium" and \
                   names[0] != "Aluminum":

                    if counts[0] == 1 and counts[1] == 1:
                        name = names[0] + "(II)" + " " + names[1]
                    elif counts[0] == 2 and counts[1] == 1:
                        name = names[0] + "(I)" + " " + names[1]
                    elif counts[0] == 1 and counts[1] == 2:
                        name = names[0] + "(IV)" + " " + names[1]
                    elif counts[0] == 2 and counts[1] == 3:
                        name = names[0] + "(III)" + " " + names[1]
                    elif counts[0] == 2 and counts[1] == 5:
                        name = names[0] + "(V)" + " " + names[1]
                    elif counts[0] == 1 and counts[1] == 3:
                        name = names[0] + "(VI)" + " " + names[1]
                    elif counts[0] == 2 and counts[1] == 7:
                        name = names[0] + "(VII)" + " " + names[1]

            if names[1] == "Nitride" or names[1] == "Phosphide":
                if names[0] != "Lithium" and names[0] != "Sodium" and names[0] != "Potassium" and \
                   names[0] != "Rubidium" and names[0] != "Cesium" and names[0] != "Francium" and \
                   names[0] != "Beryllium" and names[0] != "Magnesium" and names[0] != "Calcium" and \
                   names[0] != "Strontium" and names[0] != "Barium" and names[0] != "Radium" and \
                   names[0] != "Aluminum":

                    if counts[0] == 1 and counts[1] == 1:
                        name = names[0] + "(III)" + " " + names[1]
                    if counts[0] == 1 and counts[1] == 2:
                        name = names[0] + "(VI)" + " " + names[1]
                    elif counts[0] == 3 and counts[1] == 2:
                        name = names[0] + "(II)" + " " + names[1]
                    elif counts[0] == 3 and counts[1] == 4:
                        name = names[0] + "(IV)" + " " + names[1]
                    elif counts[0] == 3 and counts[1] == 5:
                        name = names[0] + "(V)" + " " + names[1]
                    elif counts[0] == 3 and counts[1] == 7:
                        name = names[0] + "(VII)" + " " + names[1]

            if names[1] == "Carbide":
                if names[0] != "Lithium" and names[0] != "Sodium" and names[0] != "Potassium" and \
                   names[0] != "Rubidium" and names[0] != "Cesium" and names[0] != "Francium" and \
                   names[0] != "Beryllium" and names[0] != "Magnesium" and names[0] != "Calcium" and \
                   names[0] != "Strontium" and names[0] != "Barium" and names[0] != "Radium" and \
                   names[0] != "Aluminum":

                    if counts[0] == 1 and counts[1] == 1:
                        name = names[0] + "(IV)" + " " + names[1]
                    if counts[0] == 2 and counts[1] == 1:
                        name = names[0] + "(II)" + " " + names[1]
                    if counts[0] == 4 and counts[1] == 1:
                        name = names[0] + "(I)" + " " + names[1]
                    if counts[0] == 4 and counts[1] == 3:
                        name = names[0] + "(III)" + " " + names[1]
                    if counts[0] == 4 and counts[1] == 5:
                        name = names[0] + "(V)" + " " + names[1]
                    if counts[0] == 2 and counts[1] == 3:
                        name = names[0] + "(VI)" + " " + names[1]
                    if counts[0] == 4 and counts[1] == 7:
                        name = names[0] + "(VII)" + " " + names[1]

        return name

    def symmetry(self) -> Symmetry:
        """Returns a populated Symmetry subsection."""
        result = Symmetry()
        filled = False

        if self.repr_symmetry:
            result.hall_number = self.repr_symmetry.hall_number
            result.hall_symbol = self.repr_symmetry.hall_symbol
            result.bravais_lattice = self.repr_symmetry.bravais_lattice
            result.crystal_system = self.repr_symmetry.crystal_system
            result.space_group_number = self.repr_symmetry.space_group_number
            result.space_group_symbol = self.repr_symmetry.international_short_symbol
            result.point_group = self.repr_symmetry.point_group
            filled = True

        # Fill in prototype information. SystemNormalizer has cached many of
        # the values during it's own analysis. These cached values are used
        # here.
        proto = self.repr_system.prototype if self.repr_system else None
        proto = proto[0] if proto else None
        if proto:
            # Prototype id and formula
            result.prototype_aflow_id = proto.aflow_id
            result.prototype_formula = proto.m_cache.get('prototype_name')

            # Strukturbericht: replace LaTeX with plain text
            strukturbericht = proto.m_cache.get("strukturbericht_designation")
            if strukturbericht:
                strukturbericht = re.sub('[$_{}]', '', strukturbericht)
                result.strukturbericht_designation = strukturbericht

            # Structure name. Only relevant information hidden in "notes" is
            # handed over TODO: review and eventually add more -ites which
            # are commonly used (see wurzite)
            note = proto.m_cache.get('prototype_notes')
            if note:
                structure_name_map = {
                    "CaTiO<sub>3</sub> Pnma Perovskite Structure": "perovskite",
                    "Hypothetical Tetrahedrally Bonded Carbon with 4&ndash;Member Rings": "4-member ring",
                    "In (A6) Structure": "fct",
                    "$\\alpha$&ndash;Pa (A<sub>a</sub>) Structure": "bct",
                    "Hypothetical BCT5 Si Structure": "bct5",
                    "Wurtzite (ZnS, B4) Structure": "wurtzite",
                    "Hexagonal Close Packed (Mg, A3) Structure": "hcp",
                    "Half&ndash;Heusler (C1<sub>b</sub>) Structure": "half-Heusler",
                    "Zincblende (ZnS, B3) Structure": "zincblende",
                    "Cubic Perovskite (CaTiO<sub>3</sub>, E2<sub>1</sub>) Structure": "cubic perovskite",
                    "$\\alpha$&ndash;Po (A<sub>h</sub>) Structure": "simple cubic",
                    "Si<sub>46</sub> Clathrate Structure": "clathrate",
                    "Cuprite (Cu<sub>2</sub>O, C3) Structure": "cuprite",
                    "Heusler (L2<sub>1</sub>) Structure": "Heusler",
                    "Rock Salt (NaCl, B1) Structure": "rock salt",
                    "Face&ndash;Centered Cubic (Cu, A1) Structure": "fcc",
                    "Diamond (A4) Structure": "diamond",
                    "Body&ndash;Centered Cubic (W, A2) Structure": "bcc",
                }
                result.structure_name = structure_name_map.get(note, None)

            filled = True

        if filled:
            return result
        return None



"""This tool is reading the xml file

"""
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re
import os
import os.path
import sys
from typing import Dict, Any
import xml.etree.ElementTree as ET
import numpy as np
from nomad.utils import strip
from nomad.metainfo import (
    Section, Package, SubSection, Definition, Datetime, Bytes, Unit, MEnum, Quantity)
from nomad.datamodel import EntryArchive


# URL_REGEXP from
# https://stackoverflow.com/questions/3809401/what-is-a-good-regular-expression-to-match-a-url
URL_REGEXP = re.compile(r'(https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]'
                        r'{1,256}\.[a-zA-Z0-9()]{1,6}\b([-a-zA-Z0-9()@:%_\+.~#?&//=]*))')
XML_NAMESPACES = {'nx': 'http://definition.nexusformat.org/nxdl/3.1'}

# TODO the validation still show some problems. Most notably there are a few higher
# dimensional fields with non number types, which the metainfo does not support
VALIDATE = False
CURRENT_PACKAGE: Package = None

_definition_sections: Dict[str, Section] = dict()
_XML_PARENT_MAP: Dict[ET.Element, ET.Element] = None
_NX_DOC_BASE = 'https://manual.nexusformat.org/classes'

# TODO There are more types in nxdl, but they are not used by the current base classes and
# application definitions.
_NX_TYPES = {
    'NX_FLOAT': np.dtype(np.float64),
    'NX_CHAR': str,
    'NX_BOOLEAN': bool,
    'NX_INT': np.dtype(np.int64),
    'NX_UINT': np.dtype(np.uint64),
    'NX_NUMBER': np.dtype(np.number),
    'NX_POSINT': np.dtype(np.uint64),
    'NX_BINARY': Bytes,
    'NX_DATE_TIME': Datetime
}


def to_camel_case(snake_str: str, upper: bool = False):
    """Take as input a snake case variable and return a camel case one

"""
    components = snake_str.split('_')

    if upper:
        return ''.join(f'{x[0].upper()}{x[1:]}' for x in components)

    return components[0] + ''.join(f'{x[0].upper()}{x[1:]}' for x in components[1:])


def nx_documenation_url(xml_node: ET.Element, nx_type: str):
    """Get documentation url

"""
    anchor_segments = []
    if nx_type != 'class':
        anchor_segments.append(nx_type)

    while xml_node is not None:
        if 'name' in xml_node.attrib:
            anchor_segments.append(xml_node.attrib['name'].replace('_', '-'))

        xml_node = _XML_PARENT_MAP.get(xml_node)

    anchor = "-".join([name.lower() for name in reversed(anchor_segments)])
    nx_package = CURRENT_PACKAGE.name.replace('nexus_', '')
    doc_url = f'{_NX_DOC_BASE}/{nx_package}/{anchor_segments[-1]}.html#{anchor}'
    return doc_url


def get_or_create_section(name: str, **kwargs) -> Section:
    """Returns the 'existing' metainfo section for a given top-level nexus base-class name.

    This function ensures that sections for these base-classes are only created one.
    This allows to access the metainfo section even before it is generated from the base-class
    nexus definition.

"""
    if name in _definition_sections:
        section = _definition_sections[name]
        section.more.update(**kwargs)
        return section

    section = Section(validate=VALIDATE, name=name, **kwargs)
    CURRENT_PACKAGE.section_definitions.append(section)
    _definition_sections[section.name] = section

    return section


def get_enum(xml_node: ET.Element):
    """Get the enumeration field from xml node

"""
    enumeration = xml_node.find('nx:enumeration', XML_NAMESPACES)
    if enumeration is not None:
        enum_values = []
        for enum_value in enumeration.findall('nx:item', XML_NAMESPACES):
            enum_values.append(enum_value.attrib['value'])
        return MEnum(*enum_values)
    return None


def add_common_properties_helper(base_section, definition):
    """Define definition var from base_section var

"""
    if base_section.description:
        definition.description = base_section.description
    if base_section.deprecated:
        definition.deprecated = base_section.deprecated
    if base_section.more:
        definition.more.update(**base_section.more)


def add_common_properties(xml_node: ET.Element, definition: Definition):
    '''
    Adds general metainfo definition properties (e.g. deprecated, docs, optional, ...)
    from the given nexus XML node to the given metainfo definition.
    '''
    nx_kind = definition.more.get('nx_kind')
    xml_attrs = xml_node.attrib

    # Read properties from potential base section. Those are not inherited, but we
    # duplicate them for a nicer presentation
    if isinstance(definition, Section) and definition.base_sections:
        base_section = definition.base_sections[0]
        add_common_properties_helper(base_section, definition)
    links = []
    if nx_kind is not None:
        doc_url = nx_documenation_url(xml_node, nx_kind)
        if doc_url:
            links.append(doc_url)

    doc = xml_node.find('nx:doc', XML_NAMESPACES)
    if doc is not None and doc.text is not None:
        definition.description = strip(doc.text)
        for match in URL_REGEXP.findall(definition.description):
            links.append(match[0])

    if links:
        definition.links = links

    if 'deprecated' in xml_attrs:
        definition.deprecated = xml_attrs['deprecated']

    definition.more['nx_optional'] = xml_attrs.get(
        'optional', CURRENT_PACKAGE.name == 'nexus_base_classes')

    if 'minOccurs' in xml_attrs:
        definition.more['nx_min_occurs'] = xml_attrs['minOccurs']
    if 'maxOccurs' in xml_attrs:
        definition.more['nx_max_occurs'] = xml_attrs['maxOccurs']
    if 'required' in xml_attrs:
        definition.more['nx_required'] = xml_attrs['required']
    if 'recommended' in xml_attrs:
        definition.more['nx_recommended'] = xml_attrs['recommended']

    # TODO there are probably even more nxdl attributes?


def add_attributes(xml_node: ET.Element, section: Section):
    '''
    Adds quantities for all attributes in the given nexus XML node to the given
    section.
    '''
    for attribute in xml_node.findall('nx:attribute', XML_NAMESPACES):
        attribute_section = create_attribute_section(attribute, section)

        name = attribute.attrib["name"]
        max_occurs = attribute.attrib.get('maxOccurs', '0')
        repeats = any(name_char.isupper()
                      for name_char in name) or max_occurs == 'unbounded' or int(max_occurs) > 1

        section.sub_sections.append(SubSection(
            section_def=attribute_section, nx_kind='attribute',
            name=f'nx_attribute_{name}', repeats=repeats))


def add_group_properties(xml_node: ET.Element, section: Section):
    '''
    Adds all properties that can be generated from the given nexus group XML node to
    the given (empty) metainfo section definition.
    '''
    for group in xml_node.findall('nx:group', XML_NAMESPACES):
        group_section = create_group_section(group, section)
        section.inner_section_definitions.append(group_section)

        if 'name' in group.attrib:
            name = f'nx_group_{group.attrib["name"]}'
        else:
            name = f'nx_group_{group.attrib["type"].replace("NX", "").upper()}'

        max_occurs = group.attrib.get('maxOccurs', '0')
        repeats = any(name_char.isupper()
                      for name_char in name) or max_occurs == 'unbounded' or int(max_occurs) > 1
        section.sub_sections.append(SubSection(
            section_def=group_section, nx_kind='group', name=name, repeats=repeats))

    for field in xml_node.findall('nx:field', XML_NAMESPACES):
        field_section = create_field_section(field, section)

        name = field.attrib["name"]
        max_occurs = field.attrib.get('maxOccurs', '0')
        repeats = any(name_char.isupper()
                      for name_char in name) or max_occurs == 'unbounded' or int(max_occurs) > 1

        section.sub_sections.append(SubSection(
            section_def=field_section, nx_kind='field',
            name=f'nx_field_{name}', repeats=repeats))

    add_attributes(xml_node, section)


def add_template_properties(xml_node: ET.Element, section: Section):
    '''
    Adds potential abilities of a group or field section to act as a TEMPLATE or
    nameType="any" definition.
    '''
    is_template = section.name.lower() != section.name or 'name' not in xml_node.attrib
    if is_template:
        section.quantities.append(Quantity(
            name='nx_name', type=str, default=xml_node.attrib.get('name'), description='''
                This is a nexus template property. This quantity holds the actual name used
                in the nexus data.'''))


def add_base_section(section: Section, container: Section, default_base_section: Section = None):
    '''
    Potentially adds a base section to the given section, if the given container has
    a base-section with a suitable base.
    '''
    base_section = container.all_inner_section_definitions.get(section.name, None)
    if base_section:
        assert base_section.nx_kind == section.nx_kind, 'base section has wrong nexus kind'
    else:
        base_section = default_base_section

    if base_section:
        section.base_sections = [base_section]


def create_attribute_section(xml_node: ET.Element, container: Section) -> Section:
    '''
    Creates a metainfo section from the nexus attribute given as xml node.
    '''
    xml_attrs = xml_node.attrib
    assert 'name' in xml_attrs, 'attribute has not name'

    attribute_section = Section(
        validate=VALIDATE, nx_kind='attribute',
        name=xml_attrs['name'] + 'Attribute')
    add_base_section(attribute_section, container)
    container.inner_section_definitions.append(attribute_section)

    base_value_quantity = attribute_section.all_quantities.get('nx_value')
    if base_value_quantity is None:
        value_quantity = Quantity(
            name='nx_value', description='The value for this nexus attribute')
    else:
        value_quantity = base_value_quantity.m_copy()
    attribute_section.quantities.append(value_quantity)

    enum_type = get_enum(xml_node)
    if enum_type is not None:
        value_quantity.type = enum_type

    if value_quantity.type is None:
        value_quantity.type = str

    add_common_properties(xml_node, attribute_section)
    add_template_properties(xml_node, attribute_section)

    return attribute_section


def create_field_section(xml_node: ET.Element, container: Section):
    '''
    Creates a metainfo section from the nexus field given as xml node.
    '''
    xml_attrs = xml_node.attrib

    assert 'name' in xml_attrs, 'field has not name'
    name = xml_attrs['name'] + 'Field'
    field_section = Section(validate=VALIDATE, nx_kind='field', name=name)
    add_base_section(field_section, container)
    container.inner_section_definitions.append(field_section)

    add_template_properties(xml_node, field_section)

    base_value_quantity = field_section.all_quantities.get('nx_value')
    if base_value_quantity:
        value_quantity = base_value_quantity.m_copy()
    else:
        value_quantity = Quantity(name='nx_value', description='The value for this nexus field')
    field_section.quantities.append(value_quantity)

    if 'type' in xml_attrs:
        nx_type = xml_attrs['type']
        if nx_type not in _NX_TYPES:
            raise NotImplementedError(f'type {nx_type} is not supported')
        field_section.more['nx_type'] = nx_type

        if value_quantity.type is None or value_quantity.type is Any or nx_type != 'NX_CHAR':
            value_quantity.type = _NX_TYPES[nx_type]

    enum_type = get_enum(xml_node)
    if enum_type:
        value_quantity.type = enum_type

    if value_quantity.type is None:
        value_quantity.type = Any

    if 'units' in xml_attrs:
        field_section.more['nx_units'] = xml_attrs['units']
        if xml_attrs['units'] != 'NX_UNITLESS':
            # TODO a default could be created from the nx_units value
            field_section.quantities.append(Quantity(
                name='nx_unit', type=Unit,
                description='The specific unit for that this fields data has.'))

    dimensions = xml_node.find('nx:dimensions', XML_NAMESPACES)
    if dimensions is not None:
        shape = []
        for dimension in dimensions.findall('nx:dim', XML_NAMESPACES):
            dimension_value: Any = dimension.attrib.get('value', '*')
            try:
                dimension_value = int(dimension_value)
            except ValueError:
                pass
            shape.append(dimension_value)
        value_quantity.shape = shape

    add_common_properties(xml_node, field_section)
    add_attributes(xml_node, field_section)

    return field_section


def create_group_section(xml_node: ET.Element, container: Section) -> Section:
    '''
    Creates a metainfo section from the nexus group given as xml node.
    '''
    xml_attrs = xml_node.attrib
    typ = xml_attrs['type']

    if 'name' in xml_attrs:
        name = xml_attrs['name'] + 'Group'
    else:
        name = typ + 'Group'

    group_section = Section(validate=VALIDATE, nx_kind='group', name=name)
    add_base_section(group_section, container, get_or_create_section(typ))

    add_common_properties(xml_node, group_section)
    add_template_properties(xml_node, group_section)
    add_group_properties(xml_node, group_section)

    return group_section


def create_class_section(xml_node: ET.Element) -> Section:
    '''
    Creates a metainfo section from the top-level nexus definition given as xml node.
    '''
    xml_attrs = xml_node.attrib
    assert 'name' in xml_attrs

    class_section = get_or_create_section(xml_attrs['name'], nx_kind=xml_attrs['type'])

    if 'extends' in xml_attrs:
        base_section = get_or_create_section(xml_attrs['extends'])
        class_section.base_sections = [base_section]

    add_common_properties(xml_node, class_section)
    add_group_properties(xml_node, class_section)

    return class_section


def create_package_from_nxdl_directory(path: str) -> Package:
    '''
    Creates a metainfo package from the given nexus directory. Will generate the respective
    metainfo definitions from all the nxdl files in that directory.
    '''
    global CURRENT_PACKAGE  # pylint: disable=global-statement
    CURRENT_PACKAGE = Package(name=f'nexus_{os.path.basename(path)}')

    for nxdl_file in sorted(os.listdir(path)):
        if not nxdl_file.endswith('.nxdl.xml'):
            continue

        try:
            nxdl_path = os.path.join(path, nxdl_file)
            xml_tree = ET.parse(nxdl_path)
            xml_node = xml_tree.getroot()

            global _XML_PARENT_MAP  # pylint: disable=global-statement
            _XML_PARENT_MAP = {child: parent for parent in xml_tree.iter() for child in parent}

            assert xml_node.attrib.get('type') == 'group', 'definition is not a group'

            # The section gets already implicitly added to CURRENT_PACKAGE by
            # get_or_create_section
            create_class_section(xml_node)

        except Exception as exc:
            print(f'Exception while mapping {nxdl_file}', file=sys.stderr)
            raise exc

    return CURRENT_PACKAGE


NX_DEFINITIONS_PATH = os.path.join(
    os.path.dirname(__file__),
    '../definitions')


# We generate separated metainfo package for the nexus base classes and application
# definitions.
BASE_CLASSES = create_package_from_nxdl_directory(os.path.join(NX_DEFINITIONS_PATH, 'base_classes'))
APPLICATIONS = create_package_from_nxdl_directory(os.path.join(NX_DEFINITIONS_PATH, 'applications'))
# TODO there are problems generating with nx_package='contributed_definitions'
PACKAGES = (BASE_CLASSES, APPLICATIONS)

# We take the application definitions and create a common parent section that allows to
# include nexus in an EntryArchive.
NEXUS_SECTION = Section(validate=VALIDATE, name='Nexus')

for application_section in APPLICATIONS.section_definitions:  # pylint: disable=not-an-iterable
    sub_section = SubSection(
        section_def=application_section,
        name=application_section.name.replace('NX', 'nx_application_'))
    NEXUS_SECTION.sub_sections.append(sub_section)

APPLICATIONS.section_definitions.append(NEXUS_SECTION)

ENTRY_ARCHIVE_NEXUS_SUB_SECTION = SubSection(name='nexus', section_def=NEXUS_SECTION)
EntryArchive.nexus = ENTRY_ARCHIVE_NEXUS_SUB_SECTION  # type: ignore
EntryArchive.m_def.sub_sections.append(ENTRY_ARCHIVE_NEXUS_SUB_SECTION)


# We need to initialize the metainfo definitions. This is usually done automatically,
# when the metainfo schema is defined though MSection Python classes.
for package in PACKAGES:
    package.init_metainfo()


# We skip the Python code generation for now and offer Python classes as variables
# TODO not necessary right now, could also be done case-by-case by the nexus parser
PYTHON_MODULE = sys.modules[__name__]
for package in PACKAGES:
    for sektion in package.section_definitions:  # pylint: disable=not-an-iterable
        setattr(PYTHON_MODULE, sektion.name, sektion.section_cls)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.datamodel.metainfo import simulation


m_package = Package()


class Run(simulation.run.Run):

    m_def = Section(validate=False, extends_base_section=True)

    x_fplo_t_program_version_main = Quantity(
        type=str,
        shape=[],
        description='''
        temporary: FPLO main version
        ''')

    x_fplo_t_program_version_release = Quantity(
        type=str,
        shape=[],
        description='''
        temporary: FPLO release number
        ''')

    x_fplo_t_run_hosts = Quantity(
        type=str,
        shape=[],
        description='''
        temporary: FPLO run hosts
        ''')


class System(simulation.system.System):

    m_def = Section(validate=False, extends_base_section=True)

    x_fplo_t_vec_a_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for direct lattice vectors, x-component
        ''')

    x_fplo_t_vec_a_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for direct lattice vectors, y-component
        ''')

    x_fplo_t_vec_a_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for direct lattice vectors, z-component
        ''')

    x_fplo_t_vec_b_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for reciprocal lattice vectors, x-component
        ''')

    x_fplo_t_vec_b_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for reciprocal lattice vectors, y-component
        ''')

    x_fplo_t_vec_b_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for reciprocal lattice vectors, z-component
        ''')

    x_fplo_t_atom_positions_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for atom positions, x-component
        ''')

    x_fplo_t_atom_positions_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for atom positions, y-component
        ''')

    x_fplo_t_atom_positions_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for atom positions, z-component
        ''')

    x_fplo_t_atom_idx = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Temporary storage for FPLO atom index
        ''')

    x_fplo_t_atom_labels = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storage for FPLO atom labels
        ''')

    x_fplo_t_atom_wyckoff_idx = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Temporary storage for FPLO Wyckoff position index of each atom
        ''')

    x_fplo_t_atom_cpa_block = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Temporary storage for FPLO CPA block of each atom
        ''')


class Method(simulation.method.Method):

    m_def = Section(validate=False, extends_base_section=True)

    x_fplo_t_relativity_method = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storage for FPLO relativistic method
        ''')

    x_fplo_t_dft_plus_u_species_subshell_species = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Temporary storage for FPLO per species/(n,l)subshell DFT+U species
        ''')

    x_fplo_t_dft_plus_u_species_subshell_element = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storage for FPLO per species/(n,l)subshell DFT+U element
        ''')

    x_fplo_t_dft_plus_u_species_subshell_subshell = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storage for FPLO per species/(n,l)subshell DFT+U (n,l)subshell
        ''')

    x_fplo_t_dft_plus_u_species_subshell_F0 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for FPLO per species/(n,l)subshell DFT+U F0
        ''')

    x_fplo_t_dft_plus_u_species_subshell_F2 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for FPLO per species/(n,l)subshell DFT+U F2
        ''')

    x_fplo_t_dft_plus_u_species_subshell_F4 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for FPLO per species/(n,l)subshell DFT+U F4
        ''')

    x_fplo_t_dft_plus_u_species_subshell_F6 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for FPLO per species/(n,l)subshell DFT+U F6
        ''')

    x_fplo_t_dft_plus_u_species_subshell_U = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for FPLO per species/(n,l)subshell DFT+U U
        ''')

    x_fplo_t_dft_plus_u_species_subshell_J = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for FPLO per species/(n,l)subshell DFT+U J
        ''')

    x_fplo_t_dft_plus_u_site_index = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Temporary storage for FPLO per site DFT+U index
        ''')

    x_fplo_t_dft_plus_u_site_element = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storage for FPLO per site DFT+U element
        ''')

    x_fplo_t_dft_plus_u_site_species = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Temporary storage for FPLO per site DFT+U species
        ''')

    x_fplo_t_dft_plus_u_site_subshell = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storage for FPLO per site DFT+U (n,l)subshell
        ''')

    x_fplo_t_dft_plus_u_site_ubi1 = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Temporary storage for FPLO per site DFT+U ubi1
        ''')

    x_fplo_t_dft_plus_u_site_ubi2 = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Temporary storage for FPLO per site DFT+U ubi2
        ''')


class ScfIteration(simulation.calculation.ScfIteration):

    m_def = Section(validate=False, extends_base_section=True)

    x_fplo_t_energy_reference_fermi_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Temporary storage for FPLO Fermi energy in iteration
        ''')
